import dataclasses
import json
import datetime
import tempfile
from platform import system

import pytest

from pywa import WhatsApp, types, utils, _helpers as helpers, filters
from pywa.types import Contact
from pywa.types.media import Media

PHONE_ID = "123456789"
TOKEN = "xyz"
WABA_ID = "987654321"
MSG_ID = "wamid.xx=="
MEDIA_ID = "mediaid.xx=="
TO = "135792468"
SENT_MESSAGE = {
    "messaging_product": "whatsapp",
    "contacts": [{"input": TO, "wa_id": TO}],
    "messages": [{"id": MSG_ID}],
}


@pytest.fixture
def api(mocker):
    return mocker.patch("pywa.client.WhatsApp.api")


@pytest.fixture
def wa():
    return WhatsApp(phone_id=PHONE_ID, token=TOKEN)


def test_api_usage_without_token():
    with pytest.raises(ValueError):
        _ = WhatsApp(phone_id="123", token="").api

    with pytest.raises(ValueError):
        _ = WhatsApp(token=None).api


def test_warning_when_version_lower_than_min():
    with pytest.warns(RuntimeWarning):
        WhatsApp(phone_id="123", token="123", api_version="16.0")


def test_wa_callback_scopes():
    with pytest.raises(ValueError):
        WhatsApp(
            callback_url_scope=utils.CallbackURLScope.APP,
            app_id=None,
            app_secret=None,
            server=None,
            callback_url="https://exmaple.com",
            verify_token="xyzxyz",
        )

    with pytest.raises(ValueError):
        WhatsApp(
            callback_url_scope=utils.CallbackURLScope.WABA,
            business_account_id=None,
            server=None,
            callback_url="https://exmaple.com",
            verify_token="xyzxyz",
        )

    with pytest.raises(ValueError):
        WhatsApp(
            callback_url_scope=utils.CallbackURLScope.PHONE,
            phone_id=None,
            server=None,
            callback_url="https://exmaple.com",
            verify_token="xyzxyz",
        )


def test_resolve_buttons_param():
    assert helpers.resolve_buttons_param(
        [
            types.Button(title="Button 1", callback_data="button1"),
            types.Button(title="Button 2", callback_data="button2"),
        ]
    ) == (
        types.others.InteractiveType.BUTTON,
        {
            "buttons": (
                {"type": "reply", "reply": {"id": "button1", "title": "Button 1"}},
                {"type": "reply", "reply": {"id": "button2", "title": "Button 2"}},
            )
        },
        {"_callback_options": {"button2", "button1"}},
    )

    assert helpers.resolve_buttons_param(
        types.URLButton(title="PyWa Docs", url="https://pywa.readthedocs.io")
    ) == (
        types.others.InteractiveType.CTA_URL,
        {
            "name": "cta_url",
            "parameters": {
                "display_text": "PyWa Docs",
                "url": "https://pywa.readthedocs.io",
            },
        },
        {},
    )

    assert helpers.resolve_buttons_param(
        types.VoiceCallButton(
            title="Call me",
            ttl_minutes=50,
        )
    ) == (
        types.others.InteractiveType.VOICE_CALL,
        {
            "name": "voice_call",
            "parameters": {
                "display_text": "Call me",
                "ttl_minutes": 50,
            },
        },
        {},
    )

    assert helpers.resolve_buttons_param(
        types.FlowButton(
            title="Next",
            flow_id="flow_id",
            flow_token="flow_token",
            flow_action_type=types.FlowActionType.NAVIGATE,
            flow_action_screen="START",
            flow_action_payload={"key": "value"},
            flow_message_version="3",
        )
    ) == (
        types.others.InteractiveType.FLOW,
        {
            "name": "flow",
            "parameters": {
                "mode": "published",
                "flow_message_version": "3",
                "flow_token": "flow_token",
                "flow_id": "flow_id",
                "flow_cta": "Next",
                "flow_action": "navigate",
                "flow_action_payload": {
                    "screen": "START",
                    "data": {"key": "value"},
                },
            },
        },
        {"_flow_token": "flow_token"},
    )

    assert helpers.resolve_buttons_param(
        types.SectionList(
            button_title="Menu",
            sections=[
                types.Section(
                    title="Section 1",
                    rows=[
                        types.SectionRow(
                            title="Row 1",
                            description="Description 1",
                            callback_data="row1",
                        )
                    ],
                )
            ],
        )
    ) == (
        types.others.InteractiveType.LIST,
        {
            "button": "Menu",
            "sections": (
                {
                    "title": "Section 1",
                    "rows": (
                        {
                            "id": "row1",
                            "title": "Row 1",
                            "description": "Description 1",
                        },
                    ),
                },
            ),
        },
        {"_callback_options": {"row1"}},
    )


def test_get_interactive_msg():
    assert helpers.get_interactive_msg(
        typ=types.others.InteractiveType.BUTTON,
        action={
            "buttons": (
                {"type": "reply", "reply": {"id": "button1", "title": "Button 1"}},
                {"type": "reply", "reply": {"id": "button2", "title": "Button 2"}},
            )
        },
        header={
            "type": types.MessageType.TEXT,
            "text": "header text",
        },
        body="body text",
        footer="footer text",
    ) == {
        "type": types.MessageType.BUTTON,
        "action": {
            "buttons": (
                {"type": "reply", "reply": {"id": "button1", "title": "Button 1"}},
                {"type": "reply", "reply": {"id": "button2", "title": "Button 2"}},
            )
        },
        "header": {
            "type": types.MessageType.TEXT,
            "text": "header text",
        },
        "body": {"text": "body text"},
        "footer": {"text": "footer text"},
    }


def test_get_media_msg():
    assert helpers.get_media_msg(
        media_id_or_url="1234",
        is_url=False,
        caption="caption",
        filename="filename",
    ) == {
        "id": "1234",
        "caption": "caption",
        "filename": "filename",
    }


def test_get_flow_metric_field():
    assert (
        helpers.get_flow_metric_field(
            metric_name=types.FlowMetricName.ENDPOINT_AVAILABILITY,
            granularity=types.FlowMetricGranularity.HOUR,
            since=datetime.date(2021, 1, 1),
            until=datetime.date(2021, 1, 2),
        )
        == "metric.name(ENDPOINT_AVAILABILITY).granularity(HOUR).since(2021-01-01).until(2021-01-02)"
    )

    assert (
        helpers.get_flow_metric_field(
            metric_name=types.FlowMetricName.ENDPOINT_AVAILABILITY,
            granularity=types.FlowMetricGranularity.HOUR,
            since=datetime.date(2021, 1, 1),
            until=None,
        )
        == "metric.name(ENDPOINT_AVAILABILITY).granularity(HOUR).since(2021-01-01)"
    )

    assert (
        helpers.get_flow_metric_field(
            metric_name=types.FlowMetricName.ENDPOINT_AVAILABILITY,
            granularity=types.FlowMetricGranularity.HOUR,
            since=None,
            until=datetime.date(2021, 1, 2),
        )
        == "metric.name(ENDPOINT_AVAILABILITY).granularity(HOUR).until(2021-01-02)"
    )


def test_resolve_callback_data():
    @dataclasses.dataclass
    class User(types.CallbackData):
        __callback_id__ = "user"
        __callback_data_sep__ = ":"
        id: str
        name: str

    assert helpers.resolve_callback_data(User(id="123", name="John")) == "user:123:John"

    assert helpers.resolve_callback_data("user:123:John") == "user:123:John"


def test_resolve_tracker_param():
    @dataclasses.dataclass
    class User(types.CallbackData):
        __callback_id__ = "user"
        __callback_data_sep__ = ":"
        id: str
        name: str

    assert helpers.resolve_tracker_param(User(id="123", name="John")) == "user:123:John"
    assert helpers.resolve_tracker_param("user:123:John") == "user:123:John"


def test_resolve_flow_json_param():
    assert json.loads(helpers.resolve_flow_json_param("""{"key": "value"}""")) == {
        "key": "value"
    }

    assert json.loads(helpers.resolve_flow_json_param({"key": "value"})) == {
        "key": "value"
    }

    assert json.loads(helpers.resolve_flow_json_param(b'{"key": "value"}')) == {
        "key": "value"
    }

    assert json.loads(
        helpers.resolve_flow_json_param(
            types.flows.FlowJSON(
                version="6.0",
                screens=[
                    types.flows.Screen(
                        id="START",
                        layout=types.flows.Layout(
                            children=[
                                types.flows.TextHeading(
                                    text="Hello World",
                                )
                            ],
                        ),
                    )
                ],
            )
        )
    ) == {
        "version": "6.0",
        "screens": [
            {
                "id": "START",
                "layout": {
                    "type": "SingleColumnLayout",
                    "children": [
                        {
                            "type": "TextHeading",
                            "text": "Hello World",
                        }
                    ],
                },
            }
        ],
    }


def test_resolve_waba_id_param():
    client = WhatsApp(business_account_id=WABA_ID)

    assert (
        helpers.resolve_arg(
            wa=client,
            value=None,
            method_arg="waba_id",
            client_arg="business_account_id",
        )
        == WABA_ID
    )
    assert (
        helpers.resolve_arg(
            wa=client,
            value=987654321,
            method_arg="waba_id",
            client_arg="business_account_id",
        )
        == "987654321"
    )

    with pytest.raises(ValueError):
        helpers.resolve_arg(
            wa=WhatsApp(),
            value=None,
            method_arg="waba_id",
            client_arg="business_account_id",
        )


def test_resolve_phone_id_param():
    client = WhatsApp(phone_id=PHONE_ID)

    assert (
        helpers.resolve_arg(
            wa=client, value=None, method_arg="phone_id", client_arg="phone_id"
        )
        == PHONE_ID
    )
    assert (
        helpers.resolve_arg(
            wa=client, value="1234567890", method_arg="phone_id", client_arg="phone_id"
        )
        == "1234567890"
    )

    with pytest.raises(ValueError):
        helpers.resolve_arg(
            wa=WhatsApp(), value=None, method_arg="phone_id", client_arg="phone_id"
        )


def test_resolve_media_param_with_url():
    assert helpers.resolve_media_param(
        wa=wa,
        media="https://example.com/image.jpg",
        mime_type=None,
        filename=None,
        media_type=types.MessageType.IMAGE,
        phone_id=PHONE_ID,
    ) == (True, "https://example.com/image.jpg")


def test_resolve_media_param_with_media_id():
    assert helpers.resolve_media_param(
        wa=wa,
        media="1234567890",
        mime_type=None,
        filename=None,
        media_type=types.MessageType.IMAGE,
        phone_id=PHONE_ID,
    ) == (False, "1234567890")


def test_resolve_media_param_with_media():
    assert helpers.resolve_media_param(
        wa=wa,
        media=Media(_client=wa, id="1234567890"),
        mime_type=None,
        filename=None,
        media_type=types.MessageType.IMAGE,
        phone_id=PHONE_ID,
    ) == (False, "1234567890")


def test_check_for_async_callback():
    client = WhatsApp(server=None, verify_token="xyz")

    async def callback(): ...

    with pytest.raises(ValueError):
        client._check_for_async_callback(callback)


def test_check_for_async_filters():
    client = WhatsApp(server=None, verify_token="xyz")

    async def callback(_, __) -> bool: ...

    with pytest.raises(ValueError):
        client._check_for_async_filters(filters.new(callback))


def test_token(wa):
    assert wa.token == TOKEN
    wa.token = "abc"
    assert wa.token == "abc"
    assert wa.api._session.headers["Authorization"] == "Bearer abc"


def test_send_message(api, wa):
    api.return_value.send_message.return_value = SENT_MESSAGE
    wa.send_text(
        to=TO,
        text="Hello",
        reply_to_message_id=MSG_ID,
    )
    api.send_message.assert_called_once_with(
        sender=PHONE_ID,
        to=TO,
        typ=types.MessageType.TEXT,
        msg={"body": "Hello", "preview_url": False},
        reply_to_message_id=MSG_ID,
        biz_opaque_callback_data=None,
    )


def test_send_image(api, wa):
    api.return_value.send_message.return_value = SENT_MESSAGE
    wa.send_image(
        to=TO,
        image="https://example.com/image.jpg",
        caption="caption",
        reply_to_message_id=MSG_ID,
    )
    api.send_message.assert_called_once_with(
        sender=PHONE_ID,
        to=TO,
        typ=types.MessageType.IMAGE,
        msg={
            "link": "https://example.com/image.jpg",
            "caption": "caption",
        },
        reply_to_message_id=MSG_ID,
        biz_opaque_callback_data=None,
    )


def test_send_video(api, wa):
    api.return_value.send_message.return_value = SENT_MESSAGE
    wa.send_video(
        to=TO,
        video="1234567890",
        caption="caption",
    )
    api.send_message.assert_called_once_with(
        sender=PHONE_ID,
        to=TO,
        typ=types.MessageType.VIDEO,
        msg={
            "id": "1234567890",
            "caption": "caption",
        },
        reply_to_message_id=None,
        biz_opaque_callback_data=None,
    )


def test_send_audio(api, wa):
    api.return_value.send_message.return_value = SENT_MESSAGE
    wa.send_audio(
        to=TO,
        audio="1234567890",
    )
    api.send_message.assert_called_once_with(
        sender=PHONE_ID,
        to=TO,
        typ=types.MessageType.AUDIO,
        msg={
            "id": "1234567890",
        },
        reply_to_message_id=None,
        biz_opaque_callback_data=None,
    )


def test_send_document(api, wa):
    api.return_value.send_message.return_value = SENT_MESSAGE
    wa.send_document(
        to=TO,
        document="1234567890",
        filename="filename",
    )
    api.send_message.assert_called_once_with(
        sender=PHONE_ID,
        to=TO,
        typ=types.MessageType.DOCUMENT,
        msg={
            "id": "1234567890",
            "filename": "filename",
        },
        reply_to_message_id=None,
        biz_opaque_callback_data=None,
    )


def test_send_location(api, wa):
    api.return_value.send_message.return_value = SENT_MESSAGE
    wa.send_location(
        to=TO,
        latitude=12.34,
        longitude=56.78,
        address="Address",
    )
    api.send_message.assert_called_once_with(
        sender=PHONE_ID,
        to=TO,
        typ=types.MessageType.LOCATION,
        msg={
            "latitude": 12.34,
            "longitude": 56.78,
            "name": None,
            "address": "Address",
        },
        reply_to_message_id=None,
        biz_opaque_callback_data=None,
    )


def test_send_contact(api, wa):
    api.return_value.send_message.return_value = SENT_MESSAGE
    wa.send_contact(
        to=TO,
        contact=types.Contact(
            name=Contact.Name(
                formatted_name="John Doe", first_name="John", last_name="Doe"
            ),
            phones=[Contact.Phone(phone=TO, type="WORK")],
            birthday="2000-01-01",
            emails=[Contact.Email(email="john@doe.com", type="WORK")],
            addresses=[
                Contact.Address(street="123 Main St", city="Anytown", country="USA")
            ],
            urls=[Contact.Url(url="https://example.com", type="WORK")],
            org=Contact.Org(company="Example Inc", department="Engineering"),
        ),
    )
    api.send_message.assert_called_once_with(
        sender=PHONE_ID,
        to=TO,
        typ=types.MessageType.CONTACTS,
        msg=(
            {
                "name": {
                    "formatted_name": "John Doe",
                    "first_name": "John",
                    "last_name": "Doe",
                    "middle_name": None,
                    "suffix": None,
                    "prefix": None,
                },
                "birthday": "2000-01-01",
                "phones": ({"phone": "135792468", "type": "WORK", "wa_id": None},),
                "emails": ({"email": "john@doe.com", "type": "WORK"},),
                "urls": ({"url": "https://example.com", "type": "WORK"},),
                "addresses": (
                    {
                        "street": "123 Main St",
                        "city": "Anytown",
                        "state": None,
                        "zip": None,
                        "country": "USA",
                        "country_code": None,
                        "type": None,
                    },
                ),
                "org": {
                    "company": "Example Inc",
                    "department": "Engineering",
                    "title": None,
                },
            },
        ),
        reply_to_message_id=None,
        biz_opaque_callback_data=None,
    )


def test_send_sticker(api, wa):
    api.return_value.send_message.return_value = SENT_MESSAGE
    wa.send_sticker(
        to=TO,
        sticker="1234567890",
    )
    api.send_message.assert_called_once_with(
        sender=PHONE_ID,
        to=TO,
        typ=types.MessageType.STICKER,
        msg={
            "id": "1234567890",
        },
        reply_to_message_id=None,
        biz_opaque_callback_data=None,
    )


def test_send_reaction(api, wa):
    api.return_value.send_message.return_value = SENT_MESSAGE
    wa.send_reaction(
        to=TO,
        emoji="😍",
        message_id=MSG_ID,
    )
    api.send_message.assert_called_once_with(
        sender=PHONE_ID,
        to=TO,
        typ=types.MessageType.REACTION,
        msg={
            "emoji": "😍",
            "message_id": MSG_ID,
        },
        biz_opaque_callback_data=None,
    )


def test_remove_reaction(api, wa):
    api.return_value.send_message.return_value = SENT_MESSAGE
    wa.remove_reaction(
        to=TO,
        message_id=MSG_ID,
    )
    api.send_message.assert_called_once_with(
        sender=PHONE_ID,
        to=TO,
        typ=types.MessageType.REACTION,
        msg={
            "emoji": "",
            "message_id": MSG_ID,
        },
        biz_opaque_callback_data=None,
    )


def test_request_location(api, wa):
    api.return_value.send_message.return_value = SENT_MESSAGE
    wa.request_location(
        to=TO,
        text="Send your location",
        reply_to_message_id=MSG_ID,
    )
    api.send_message.assert_called_once_with(
        sender=PHONE_ID,
        to=TO,
        typ=types.MessageType.INTERACTIVE,
        msg={
            "action": {
                "name": "send_location",
            },
            "body": {"text": "Send your location"},
            "type": types.others.InteractiveType.LOCATION_REQUEST_MESSAGE,
        },
        reply_to_message_id=MSG_ID,
        biz_opaque_callback_data=None,
    )


def test_send_catalog(api, wa):
    api.return_value.send_message.return_value = SENT_MESSAGE
    wa.send_catalog(
        to=TO,
        body="Hello",
        thumbnail_product_sku="IPHONE_8",
    )
    api.send_message.assert_called_once_with(
        sender=PHONE_ID,
        to=TO,
        typ=types.MessageType.INTERACTIVE,
        msg={
            "action": {
                "name": "catalog_message",
                "parameters": {
                    "thumbnail_product_retailer_id": "IPHONE_8",
                },
            },
            "body": {"text": "Hello"},
            "type": types.others.InteractiveType.CATALOG_MESSAGE,
        },
        reply_to_message_id=None,
        biz_opaque_callback_data=None,
    )


def test_send_product(api, wa):
    api.return_value.send_message.return_value = SENT_MESSAGE
    wa.send_product(
        to=TO,
        catalog_id="1234567890",
        sku="IPHONE_8",
    )
    api.send_message.assert_called_once_with(
        sender=PHONE_ID,
        to=TO,
        typ=types.MessageType.INTERACTIVE,
        msg={
            "action": {
                "catalog_id": "1234567890",
                "product_retailer_id": "IPHONE_8",
            },
            "type": types.others.InteractiveType.PRODUCT,
        },
        reply_to_message_id=None,
        biz_opaque_callback_data=None,
    )


def test_send_products(api, wa):
    api.return_value.send_message.return_value = SENT_MESSAGE
    wa.send_products(
        to=TO,
        title="Products",
        body="Hello",
        catalog_id="1234567890",
        product_sections=[
            types.ProductsSection(
                title="Section 1",
                skus=["IPHONE_8", "IPHONE_X"],
            )
        ],
    )
    api.send_message.assert_called_once_with(
        sender=PHONE_ID,
        to=TO,
        typ=types.MessageType.INTERACTIVE,
        msg={
            "action": {
                "catalog_id": "1234567890",
                "sections": (
                    {
                        "product_items": (
                            {
                                "product_retailer_id": "IPHONE_8",
                            },
                            {
                                "product_retailer_id": "IPHONE_X",
                            },
                        ),
                        "title": "Section 1",
                    },
                ),
            },
            "body": {"text": "Hello"},
            "header": {"text": "Products", "type": types.MessageType.TEXT},
            "type": types.others.InteractiveType.PRODUCT_LIST,
        },
        reply_to_message_id=None,
        biz_opaque_callback_data=None,
    )



from __future__ import annotations

__all__ = [
    "resolve_buttons_param",
    "resolve_media_param",
    "resolve_tracker_param",
    "resolve_arg",
    "upload_template_media_components",
    "resolve_flow_json_param",
    "get_interactive_msg",
    "get_media_msg",
    "get_flow_metric_field",
    "resolve_callback_data",
]

import datetime
import json
import mimetypes
import pathlib
import threading
from concurrent import futures
from typing import Any, BinaryIO, Literal, Iterable, TYPE_CHECKING

import httpx

from .types import (
    FlowMetricName,
    FlowMetricGranularity,
    FlowJSON,
    CallbackData,
    MessageType,
    URLButton,
    VoiceCallButton,
    SectionList,
    FlowButton,
    Button,
    CallPermissionRequestButton,
)
from pywa.types.others import InteractiveType
from .types.media import Media
from .types.templates import (
    TemplateBaseComponent,
    _BaseMediaHeaderComponent,
    HeaderFormatType,
    Carousel,
    _BaseMediaParams,
)

if TYPE_CHECKING:
    from pywa import WhatsApp


def resolve_buttons_param(
    buttons: (
        Iterable[Button]
        | URLButton
        | VoiceCallButton
        | CallPermissionRequestButton
        | FlowButton
        | SectionList
    ),
) -> tuple[
    InteractiveType,
    dict,
    dict[str, set[str] | str],
]:
    """
    Internal method to resolve ``buttons`` parameter. Returns a tuple of (``type``, ``buttons``, ``callback_options``).
    """
    if isinstance(buttons, SectionList):
        data = buttons.to_dict()
        return (
            InteractiveType.LIST,
            data,
            {
                "_callback_options": {
                    row["id"] for section in data["sections"] for row in section["rows"]
                }
            },
        )
    elif isinstance(buttons, URLButton):
        return InteractiveType.CTA_URL, buttons.to_dict(), {}
    elif isinstance(buttons, VoiceCallButton):
        return InteractiveType.VOICE_CALL, buttons.to_dict(), {}
    elif isinstance(buttons, CallPermissionRequestButton):
        return InteractiveType.CALL_PERMISSION_REQUEST, buttons.to_dict(), {}
    elif isinstance(buttons, FlowButton):
        return (
            InteractiveType.FLOW,
            buttons.to_dict(),
            {"_flow_token": buttons.flow_token},
        )
    else:  # assume its list of buttons
        data = tuple(b.to_dict() for b in buttons)
        return (
            InteractiveType.BUTTON,
            {"buttons": data},
            {"_callback_options": {b["reply"]["id"] for b in data}},
        )


_media_types_default_filenames = {
    "IMAGE": "pywa-image.jpg",
    "VIDEO": "pywa-video.mp4",
    "AUDIO": "pywa-audio.mp3",
    "STICKER": "pywa-sticker.webp",
}
_template_header_formats_filename = {
    HeaderFormatType.IMAGE: "pywa-template-header-image.jpg",
    HeaderFormatType.VIDEO: "pywa-template-header-video.mp4",
    HeaderFormatType.DOCUMENT: "pywa-template-header-document.pdf",
}
_template_header_formats_default_mime_types = {
    HeaderFormatType.IMAGE: "image/jpeg",
    HeaderFormatType.VIDEO: "video/mp4",
    HeaderFormatType.DOCUMENT: "application/pdf",
}


def resolve_media_param(
    wa: WhatsApp,
    media: str | Media | pathlib.Path | bytes | BinaryIO,
    mime_type: str | None,
    filename: str | None,
    media_type: str | None,
    phone_id: str,
) -> tuple[bool, str]:
    """
    Internal method to resolve the ``media`` parameter. Returns a tuple of (``is_url``, ``media_id_or_url``).
    """
    if isinstance(media, Media):
        return False, media.id
    if isinstance(media, (str, pathlib.Path)):
        if str(media).startswith(("https://", "http://")):
            return True, media
        elif str(media).isdigit() and not pathlib.Path(media).is_file():
            return False, media  # assume it's a media ID
    # assume its bytes or a file path
    return False, wa.upload_media(
        phone_id=phone_id,
        media=media,
        mime_type=mime_type,
        filename=_media_types_default_filenames.get(media_type, filename),
    ).id


def upload_template_media_components(
    *,
    wa: WhatsApp,
    app_id: int | str | None,
    components: list[TemplateBaseComponent | dict],
) -> None:
    """
    Internal method to upload media components examples in a template.
    """
    not_uploaded = []
    for comp in components:
        if isinstance(comp, _BaseMediaHeaderComponent) and comp._handle is None:
            not_uploaded.append(comp)
        elif isinstance(comp, Carousel):
            for card in comp.cards:
                for cc in card.components:
                    if isinstance(cc, _BaseMediaHeaderComponent) and cc._handle is None:
                        not_uploaded.append(cc)
    if not not_uploaded:
        return

    stop_event = threading.Event()
    with futures.ThreadPoolExecutor(
        thread_name_prefix="pywa-upload-template-media"
    ) as executor:
        tasks = [
            executor.submit(
                _upload_comp,
                wa=wa,
                c=c,
                app_id=app_id,
                stop_event=stop_event,
            )
            for c in not_uploaded
        ]
        for future in futures.as_completed(tasks):
            future.result()
            if stop_event.is_set():
                break


def _upload_comp(
    *,
    wa: WhatsApp,
    c: _BaseMediaHeaderComponent,
    app_id: int | str | None,
    stop_event: threading.Event,
) -> None:
    if stop_event.is_set():
        return
    example = c.example
    filename: str | None = None
    mime_type: str | None = None
    raw_bytes: bytes | None = None
    try:
        if isinstance(example, (str, pathlib.Path, Media)):
            if str(example).startswith(("https://", "http://")):  # URL
                res = httpx.Client(follow_redirects=True).get(str(example))
                res.raise_for_status()
                raw_bytes = res.content
                mime_type = res.headers.get("Content-Type")
            elif (path := pathlib.Path(example)).is_file():  # File path
                with open(path, "rb") as f:
                    filename = path.name
                    mime_type = mimetypes.guess_type(path)[0]
                    raw_bytes = f.read()
            elif str(example).isdigit() or isinstance(
                example, Media
            ):  # WhatsApp Media ID
                url_res = wa.get_media_url(
                    media_id=example.id if isinstance(example, Media) else example
                )
                mime_type = url_res.mime_type
                raw_bytes = wa.download_media(url=url_res.url, in_memory=True)
        elif isinstance(example, bytes):  # Raw bytes
            raw_bytes = example
        if not raw_bytes:
            raise ValueError(
                f"Invalid media example for component {c.__class__.__name__}: {example}. "
                "It must be a URL, file path, WhatsApp Media, or bytes."
            )
        app_id = resolve_arg(
            wa=wa,
            value=app_id,
            method_arg="app_id",
            client_arg="app_id",
        )
        c._handle = wa.api.upload_file(
            upload_session_id=wa.api.create_upload_session(
                app_id=app_id,
                file_name=filename
                or _template_header_formats_filename.get(
                    c.format, "pywa-template-header"
                ),
                file_length=len(raw_bytes),
                file_type=mime_type
                or _template_header_formats_default_mime_types.get(
                    c.format, "application/octet-stream"
                ),
            )["id"],
            file=raw_bytes,
            file_offset=0,
        )["h"]
    except Exception as e:
        stop_event.set()
        e.add_note(
            f"Failed to upload media for component {c.__class__.__name__} with example: {example}"
        )
        raise


def upload_template_media_params(
    *,
    wa: WhatsApp,
    sender: str,
    params: list[TemplateBaseComponent.Params | dict],
) -> None:
    """
    Internal method to upload media parameters when sending a template message.
    """
    not_uploaded = []
    for param in params:
        if isinstance(param, _BaseMediaParams) and param._resolved_media is None:
            not_uploaded.append(param)
        elif isinstance(param, Carousel.Params):
            for card_params in param.cards:
                for p in card_params.params:
                    if isinstance(p, _BaseMediaParams) and p._resolved_media is None:
                        not_uploaded.append(p)

    if not not_uploaded:
        return
    with futures.ThreadPoolExecutor(
        thread_name_prefix="pywa-upload-template-media-params"
    ) as executor:
        tasks = [
            executor.submit(
                _upload_param,
                wa=wa,
                sender=sender,
                p=param,
            )
            for param in not_uploaded
        ]
        for task in futures.wait(tasks)[0]:
            task.result()


def _upload_param(*, wa: WhatsApp, sender: str, p: _BaseMediaParams) -> None:
    try:
        p._is_url, p._resolved_media = resolve_media_param(
            wa=wa,
            media=p.media,
            mime_type=None,
            filename=None,
            media_type=p.format.value,
            phone_id=sender,
        )
    except Exception as e:
        e.add_note(f"Failed to upload media parameter with media: {p.media}")
        raise e


def resolve_tracker_param(tracker: str | CallbackData | None) -> str | None:
    """Internal method to resolve the `tracker` parameter."""
    return tracker.to_str() if isinstance(tracker, CallbackData) else tracker


def resolve_arg(
    *,
    wa: WhatsApp,
    value: str | int | None,
    method_arg: str,
    client_arg: str,
) -> str:
    if value is not None:
        return str(value)
    if getattr(wa, client_arg, None) is not None:
        return str(getattr(wa, client_arg))
    raise ValueError(
        f"When initializing WhatsApp without `{client_arg}`, `{method_arg}` must be provided."
    )


def resolve_flow_json_param(
    flow_json: FlowJSON | dict | str | pathlib.Path | bytes | BinaryIO,
) -> str:
    """Internal method to solve the `flow_json` parameter"""
    json_str, to_dump = None, None
    if isinstance(flow_json, (str, pathlib.Path)):  # json str or path to json file
        as_path = pathlib.Path(flow_json)
        try:
            if as_path.is_file():
                with open(as_path, "r", encoding="utf-8") as f:
                    json_str = f.read()
            else:
                json_str = flow_json
        except OSError:
            json_str = flow_json
    elif isinstance(flow_json, bytes):
        json_str = flow_json.decode()
    elif isinstance(flow_json, FlowJSON):
        json_str = flow_json.to_json()
    elif isinstance(flow_json, dict):
        to_dump = flow_json
    else:
        raise TypeError(
            f"`flow_json` must be a FlowJSON object, dict, json string, json file path or json bytes. not {type(flow_json)}"
        )

    if to_dump is not None:
        json_str = json.dumps(to_dump, indent=4, ensure_ascii=False)

    return json_str


def get_interactive_msg(
    typ: InteractiveType,
    action: dict[str, Any],
    header: dict | None = None,
    body: str | None = None,
    footer: str | None = None,
):
    return {
        "type": typ,
        "action": action,
        **({"header": header} if header else {}),
        **({"body": {"text": body}} if body else {}),
        **({"footer": {"text": footer}} if footer else {}),
    }


def get_media_msg(
    media_id_or_url: str,
    is_url: bool,
    caption: str | None = None,
    filename: str | None = None,
):
    return {
        ("link" if is_url else "id"): media_id_or_url,
        **({"caption": caption} if caption else {}),
        **({"filename": filename} if filename else {}),
    }


def get_flow_metric_field(
    metric_name: FlowMetricName,
    granularity: FlowMetricGranularity,
    since: datetime.date | str | None,
    until: datetime.date | str | None,
) -> str:
    date_fmt = "%Y-%m-%d"
    return (
        f"metric.name({metric_name}).granularity({granularity})"
        + (
            f".since({since.strftime(date_fmt) if isinstance(since, datetime.date) else since})"
            if since
            else ""
        )
        + (
            f".until({until.strftime(date_fmt) if isinstance(until, datetime.date) else until})"
            if until
            else ""
        )
    )


def resolve_callback_data(data: str | CallbackData) -> str:
    """Internal function to convert callback data to a string."""
    if isinstance(data, CallbackData):
        return data.to_str()
    elif isinstance(data, str):
        return data
    raise TypeError(f"Invalid callback data type {type(data)}")



"""This module contains the types related to message status updates."""

from __future__ import annotations

import warnings

from .callback import _CallbackDataT

__all__ = [
    "MessageStatus",
    "MessageStatusType",
    "Conversation",
    "ConversationCategory",
    "Pricing",
    "PricingModel",
    "PricingType",
    "PricingCategory",
]

import dataclasses
import logging
import datetime
from typing import TYPE_CHECKING, Generic

from .. import utils
from ..errors import WhatsAppError

from .base_update import BaseUserUpdate  # noqa
from .others import Metadata, User

if TYPE_CHECKING:
    from ..client import WhatsApp

_logger = logging.getLogger(__name__)


class MessageStatusType(utils.StrEnum):
    """
    Message status type.

    Attributes:
        SENT: The message was sent.
        DELIVERED: The message was delivered.
        READ: The message was read.
        FAILED: The message failed to send (you can access the ``.error`` attribute for more information).
    """

    _check_value = str.islower
    _modify_value = str.lower

    SENT = "sent"
    DELIVERED = "delivered"
    READ = "read"
    FAILED = "failed"

    UNKNOWN = "UNKNOWN"


class ConversationCategory(utils.StrEnum):
    """
    Conversation category.

    Attributes:
        AUTHENTICATION: The conversation is related to authentication.
        MARKETING: The conversation is related to marketing.
        MARKETING_LITE: The conversation is related to marketing lite.
        UTILITY: The conversation is related to utility.
        SERVICE: The conversation is related to service.
        REFERRAL_CONVERSION: The conversation is related to referral conversion.
        UNKNOWN: The conversation category is unknown.
    """

    _check_value = str.islower
    _modify_value = str.lower

    AUTHENTICATION = "authentication"
    MARKETING = "marketing"
    MARKETING_LITE = "marketing_lite"
    UTILITY = "utility"
    SERVICE = "service"
    REFERRAL_CONVERSION = "referral_conversion"

    UNKNOWN = "UNKNOWN"


@dataclasses.dataclass(frozen=True, slots=True, kw_only=True)
class MessageStatus(BaseUserUpdate, Generic[_CallbackDataT]):
    """
    Represents the status of a message.

    - `'MessageStatus' on developers.facebook.com <https://developers.facebook.com/docs/whatsapp/cloud-api/webhooks/components#statuses-object>`_.

    ``MessageStatus`` is a generic class, so when providing a ``factory`` parameter in callback handlers, you can
    specify the type of the factory to get autocomplete in the ``tracker`` attribute.

    Here is an example:

        >>> from pywa.types import CallbackData
        >>> from dataclasses import dataclass
        >>> @dataclass(frozen=True, slots=True)
        >>> class UserData(CallbackData): # Subclass CallbackData
        ...     id: int
        ...     name: str
        ...     admin: bool

        >>> from pywa import WhatsApp
        >>> from pywa.types import Button, CallbackButton
        >>> wa = WhatsApp(...)
        >>> wa.send_message(
        ...     to='972987654321',
        ...     text='Hi user',
        ...     tracker=UserData(id=123, name='david', admin=True) # Here ^^^ we use the UserData class as the tracker
        ... )           # Here ^^^ we use the UserData class as the tracker data

        >>> @wa.on_message_status(factory=UserData) # Use the factory parameter to convert the tracker data
        ... def on_status(_: WhatsApp, s: MessageStatus[UserData]): # For autocomplete
        ...    if s.tracker.admin: print(s.tracker.id) # Access the tracker data



    Attributes:
        id: The ID of the message that the status is for.
        metadata: The metadata of the message (to which phone number it was sent).
        status: The status of the message.
        timestamp: The timestamp when the status was updated (in UTC).
        from_user: The user who the message was sent to.
        conversation: The conversation that the message was sent in (See `Conversation <https://developers.facebook.com/docs/whatsapp/pricing#conversations>`_).
        pricing: The pricing of the message (Optional).
        error: The error that occurred (if status is :class:`MessageStatusType.FAILED`).
        tracker: The tracker that the message was sent with (e.g. ``wa.send_message(tracker=...)``).
        shared_data: Shared data between handlers.
    """

    status: MessageStatusType
    pricing: Pricing | None
    conversation: Conversation | None
    error: WhatsAppError | None
    tracker: _CallbackDataT | None

    _txt_fields = ("tracker",)
    _webhook_field = "messages"

    @classmethod
    def from_update(cls, client: WhatsApp, update: dict) -> MessageStatus:
        status = (value := (entry := update["entry"][0])["changes"][0]["value"])[
            "statuses"
        ][0]
        error = value.get("errors", status.get("errors", (None,)))[0]
        return cls(
            _client=client,
            raw=update,
            waba_id=entry["id"],
            id=status["id"],
            metadata=Metadata.from_dict(value["metadata"]),
            status=MessageStatusType(status["status"]),
            timestamp=datetime.datetime.fromtimestamp(
                int(status["timestamp"]),
                datetime.timezone.utc,
            ),
            from_user=client._usr_cls(
                wa_id=status["recipient_id"], name=None, _client=client
            ),
            tracker=status.get("biz_opaque_callback_data"),
            conversation=Conversation.from_dict(status["conversation"])
            if "conversation" in status
            else None,
            pricing=Pricing.from_dict(status["pricing"])
            if "pricing" in status
            else None,
            error=WhatsAppError.from_dict(error=error) if error else None,
        )


@dataclasses.dataclass(frozen=True, slots=True)
class Conversation:
    """
    Conversations are 24-hour message threads between you and your customers.
    They are opened and charged when messages you send to customers are delivered.

    - `'Conversation' on developers.facebook.com <https://developers.facebook.com/docs/whatsapp/pricing#conversations>`_.

    Attributes:
        id: Represents the ID of the conversation the given status notification belongs to.
        category: The category of the conversation.
        expiration: The expiration date (in UTC) of the conversation (Optional, only for `sent` updates).
    """

    id: str
    category: ConversationCategory
    expiration: datetime.datetime | None

    @classmethod
    def from_dict(cls, data: dict):
        return cls(
            id=data["id"],
            category=ConversationCategory(data["origin"]["type"]),
            expiration=datetime.datetime.fromtimestamp(
                int(data["expiration_timestamp"]),
                datetime.timezone.utc,
            )
            if "expiration_timestamp" in data
            else None,
        )


class PricingModel(utils.StrEnum):
    """
    Pricing model.

    Attributes:
        CBP: Indicates conversation-based pricing applies.
        PMP: Indicates per-message pricing applies.
    """

    CBP = "CBP"
    PMP = "PMP"

    UNKNOWN = "UNKNOWN"


class PricingType(utils.StrEnum):
    """
    Pricing type.

    Attributes:
        REGULAR: Indicates the message is billable.
        FREE_CUSTOMER_SERVICE: Indicates the message is free because it was either a utility template message or non-template message sent within a `customer service window <https://developers.facebook.com/docs/whatsapp/cloud-api/guides/send-messages#customer-service-windows>`_.
        FREE_ENTRY_POINT: Indicates the message is free because it is part of a `free-entry point conversation <https://developers.facebook.com/docs/whatsapp/pricing#free-entry-point-conversations>`_.
    """

    _check_value = str.islower
    _modify_value = str.lower

    REGULAR = "regular"
    FREE_CUSTOMER_SERVICE = "free_customer_service"
    FREE_ENTRY_POINT = "free_entry_point"

    UNKNOWN = "UNKNOWN"


class PricingCategory(utils.StrEnum):
    """
    Pricing category.

    Attributes:
        AUTHENTICATION: Indicates an authentication template message.
        AUTHENTICATION_INTERNATIONAL: Indicates an authentication template message sent to a WhatsApp user in a country or region that has authentication-international rates.
        MARKETING: Indicates a marketing template message.
        MARKETING_LITE: Indicates a marketing template message sent via MM Lite API.
        UTILITY: Indicates a utility template message.
        SERVICE: Indicates a non-template message.
        REFERRAL_CONVERSION: Indicates the message is part of a `free entry point conversation <https://developers.facebook.com/docs/whatsapp/pricing#free-entry-point-conversations>`_.
    """

    _check_value = str.islower
    _modify_value = str.lower

    AUTHENTICATION = "authentication"
    AUTHENTICATION_INTERNATIONAL = "authentication_international"
    MARKETING = "marketing"
    MARKETING_LITE = "marketing_lite"
    UTILITY = "utility"
    SERVICE = "service"
    REFERRAL_CONVERSION = "referral_conversion"

    UNKNOWN = "UNKNOWN"


@dataclasses.dataclass(frozen=True, slots=True)
class Pricing:
    """
    Represents the pricing of a message.

    - `'Pricing' on developers.facebook.com <https://developers.facebook.com/docs/whatsapp/pricing>`_.

    Attributes:
        billable: Indicates if the message is billable.
        model: The pricing model used for the message.
        type: The pricing type of the message (Only available from webhook v24.0^).
        category: The pricing category of the message.

    """

    billable: bool
    model: PricingModel
    type: PricingType | None
    category: PricingCategory

    @classmethod
    def from_dict(cls, data: dict):
        return cls(
            billable=data.get("billable", data.get("type") == PricingType.REGULAR),
            model=PricingModel(data["pricing_model"]),
            type=PricingType(data["pricing_type"]) if "pricing_type" in data else None,
            category=PricingCategory(data["category"]),
        )



"""This module contains the callback types."""

__all__ = [
    "CallbackButton",
    "CallbackSelection",
    "Button",
    "URLButton",
    "ButtonUrl",  #
    "VoiceCallButton",
    "CallPermissionRequestButton",
    "SectionRow",
    "Section",
    "SectionList",
    "FlowButton",
    "CallbackData",
]

from pywa.types.callback import *  # noqa MUST BE IMPORTED FIRST
from pywa.types.callback import (
    CallbackButton as _CallbackButton,
    CallbackSelection as _CallbackSelection,
    _CallbackDataT,
)  # noqa MUST BE IMPORTED FIRST
from .base_update import BaseUserUpdateAsync


import dataclasses
from typing import TYPE_CHECKING


if TYPE_CHECKING:
    from ..client import WhatsApp


@dataclasses.dataclass(frozen=True, slots=True, kw_only=True)
class CallbackButton(BaseUserUpdateAsync, _CallbackButton[_CallbackDataT]):
    """
    Represents a callback button (Incoming update when user clicks on :class:`Button` or chooses
    :class:`Template.QuickReplyButtonData`).

    ``CallbackButton`` is a generic class, so when providing a ``factory`` parameter in callback handlers, you can
    specify the type of the factory to get autocomplete in the ``data`` attribute.

    Here is an example:

        >>> from pywa.types import CallbackData
        >>> from dataclasses import dataclass
        >>> @dataclass(frozen=True, slots=True)
        >>> class UserData(CallbackData): # Subclass CallbackData
        ...     id: int
        ...     name: str
        ...     admin: bool

        >>> from pywa import WhatsApp
        >>> from pywa.types import Button, CallbackButton
        >>> wa = WhatsApp(...)
        >>> wa.send_message(
        ...     to='972987654321',
        ...     text='Click the button to get the user',
        ...     buttons=[Button(title='Get user', callback_data=UserData(id=123, name='david', admin=True))]
        ... )                                     # Here ^^^ we use the UserData class as the callback data

        >>> @wa.on_callback_button(factory=UserData) # Use the factory parameter to convert the callback data
        ... def on_user_data(_: WhatsApp, btn: CallbackButton[UserData]): # For autocomplete
        ...    if btn.data.admin: print(btn.data.id) # Access the data object as an attribute


    Attributes:
        id: The ID of the message.
        metadata: The metadata of the message (to which phone number it was sent).
        type: The message type (:class:`MessageType.INTERACTIVE` for :class:`~pywa.types.callback.Button` presses or :class:`MessageType.BUTTON` for :class:`~pywa.types.templates.QuickReplyButton` clicks).
        from_user: The user who sent the message.
        timestamp: The timestamp when the message was sent (in UTC).
        reply_to_message: The message to which this callback button is a reply to.
        data: The data of the button (the ``callback_data`` parameter you provided in :class:`Button` or
         :class:`Template.QuickReplyButtonData`).
        title: The title of the button.
        shared_data: Shared data between handlers.
    """


@dataclasses.dataclass(frozen=True, slots=True, kw_only=True)
class CallbackSelection(BaseUserUpdateAsync, _CallbackSelection[_CallbackDataT]):
    """
    Represents a callback selection (Incoming update when user clicks on :class:`SectionRow` in :class:`SectionList`).

    ``CallbackSelection`` is a generic class, so when providing a ``factory`` parameter in callback handlers, you can
    specify the type of the factory to get autocomplete in the ``data`` attribute.

    Here is an example:

        >>> from pywa.types import CallbackData
        >>> from dataclasses import dataclass
        >>> @dataclass(frozen=True, slots=True)
        >>> class UserData(CallbackData): # Subclass CallbackData
        ...     id: int
        ...     name: str
        ...     admin: bool

        >>> from pywa import WhatsApp
        >>> from pywa.types import SectionList, Section, SectionRow, CallbackSelection
        >>> wa = WhatsApp(...)
        >>> wa.send_message(
        ...     to='972987654321',
        ...     text='Click the button to get the user',
        ...     buttons=SectionList(
        ...         button_title='Get user', sections=[
        ...             Section(title='Users', rows=[
        ...                 SectionRow(title='Get user', callback_data=UserData(id=123, name='david', admin=True))
        ...             ])                              # Here ^^^ we use the UserData class as the callback data
        ...         ]
        ...     )
        ... )

        >>> @wa.on_callback_selection(factory=UserData) # Use the factory parameter to convert the callback data
        ... def on_user_data(_: WhatsApp, sel: CallbackSelection[UserData]): # For autocomplete
        ...    if sel.data.admin: print(sel.data.id) # Access the data object as an attribute

    Attributes:
        id: The ID of the message.
        metadata: The metadata of the message (to which phone number it was sent).
        type: The message type (always :class:`MessageType.INTERACTIVE`).
        from_user: The user who sent the message.
        timestamp: The timestamp when the message was sent (in UTC).
        reply_to_message: The message to which this callback selection is a reply to.
        data: The data of the selection (the ``callback_data`` parameter you provided in :class:`SectionRow`).
        title: The title of the selection.
        description: The description of the selection (optional).
    """



"""This module contains the types related to WhatsApp Flows."""

from __future__ import annotations

__all__ = [
    "FlowCompletion",
    "FlowRequest",
    "FlowResponse",
    "FlowResponseError",
    "FlowRequestCannotBeDecrypted",
    "FlowRequestSignatureAuthenticationFailed",
    "FlowTokenNoLongerValid",
    "FlowCategory",
    "FlowDetails",
    "FlowMetricName",
    "FlowMetricGranularity",
    "FlowStatus",
    "FlowPreview",
    "FlowJSONUpdateResult",
    "FlowValidationError",
    "FlowAsset",
    "CreatedFlow",
    "MigratedFlow",
    "MigratedFlowError",
    "MigrateFlowsResponse",
    "FlowJSON",
    "Screen",
    "ScreenData",
    "ScreenDataUpdate",
    "Layout",
    "LayoutType",
    "Form",
    "ScreenDataRef",
    "ComponentRef",
    "FlowStr",
    "TextHeading",
    "TextSubheading",
    "TextBody",
    "TextCaption",
    "RichText",
    "FontWeight",
    "TextInput",
    "InputType",
    "LabelVariant",
    "TextArea",
    "CheckboxGroup",
    "ChipsSelector",
    "RadioButtonsGroup",
    "Footer",
    "OptIn",
    "Dropdown",
    "EmbeddedLink",
    "NavigationList",
    "NavigationItem",
    "NavigationItemStart",
    "NavigationItemMainContent",
    "NavigationItemEnd",
    "DatePicker",
    "CalendarPicker",
    "CalendarPickerMode",
    "CalendarRangeValues",
    "CalendarDay",
    "Image",
    "ImageCarouselItem",
    "ImageCarousel",
    "PhotoPicker",
    "PhotoSource",
    "DocumentPicker",
    "ScaleType",
    "If",
    "Switch",
    "DataSource",
    "DataExchangeAction",
    "NavigateAction",
    "CompleteAction",
    "UpdateDataAction",
    "OpenUrlAction",
    "FlowActionType",
    "FlowRequestActionType",
    "Next",
    "NextType",
]

import httpx

from pywa.types.flows import *  # noqa MUST BE IMPORTED FIRST
from pywa.types.flows import (
    FlowDetails as _FlowDetails,
    FlowCompletion as _FlowCompletion,
    FlowRequest as _FlowRequest,
)  # noqa MUST BE IMPORTED FIRST
from pywa.types.others import SuccessResult
from .others import Result
from .base_update import BaseUserUpdateAsync

import dataclasses
import pathlib
from typing import Iterable, TYPE_CHECKING, BinaryIO

from .. import utils

if TYPE_CHECKING:
    from ..client import WhatsApp


@dataclasses.dataclass(slots=True, kw_only=True, frozen=True)
class FlowRequest(_FlowRequest):
    """
    Represents a flow request. This request is sent to the flow endpoint when a user interacts with a flow.

    - Read more at `developers.facebook.com <https://developers.facebook.com/docs/whatsapp/flows/guides/implementingyourflowendpoint#data_exchange_request>`_.

    Attributes:
        version: The version of the ``data_api_version`` specified on the flow json.
        flow_token: The flow token used to create the flow
        action: The action that triggered the request.
        screen: The screen that triggered the request. If action is ``FlowRequestActionType.INIT`` or ``FlowRequestActionType.BACK``, this field may be ``None``.
        data: The data sent from the screen. If action is ``FlowRequestActionType.BACK`` or ``FlowRequestActionType.INIT``, this field may be ``None``.
        raw: The raw data of the request.
        raw_encrypted: The raw-encrypted data of the request.
    """

    async def decrypt_media(
        self, key: str, index: int = 0, dl_session: httpx.AsyncClient | None = None
    ) -> utils.FlowRequestDecryptedMedia:
        """
        Decrypt the encrypted media file from the flow request.

        Example:

            >>> from pywa_async import WhatsApp, types
            >>> wa = WhatsApp(...)
            >>> @wa.on_flow_request("/my-flow-endpoint")
            ... async def my_flow_endpoint(_: WhatsApp, req: types.FlowRequest):
            ...     media_id, filename, decrypted_data = await req.decrypt_media(key="driver_license", index=0)
            ...     with open(filename, "wb") as file:
            ...         file.write(decrypted_data)
            ...     return req.respond(...)

        Args:
            key: The key of the media in the data (e.g. ``"driver_license"``).
            index: The index of the media in the data (default to ``0``).
            dl_session: The HTTPX async session to download the media (optional, new session will be created if not provided).

        Returns:
            A tuple of (media_id, filename, decrypted_data) where:
                - media_id: The media ID of the decrypted media.
                - filename: The filename of the decrypted media.
                - decrypted_data: The decrypted data of the media.
        Raises:
            ValueError: If the request has no data.
            KeyError: If the key is not found in the data.
            IndexError: If the index is out of range.
        """
        return await utils.flow_request_media_decryptor(
            encrypted_media=self.data[key][index],
            dl_session=dl_session,
        )


@dataclasses.dataclass(slots=True, kw_only=True, frozen=True)
class FlowCompletion(BaseUserUpdateAsync, _FlowCompletion):
    """
    A flow completion message. This update arrives when a user completes a flow.

    - Read more at `developers.facebook.com <https://developers.facebook.com/docs/whatsapp/flows/gettingstarted/receiveflowresponse>`_.

    Attributes:
        id: The message ID (If you want to reply to the message, use ``message_id_to_reply`` instead).
        metadata: The metadata of the message (to which phone number it was sent).
        type: The message type (Always ``MessageType.INTERACTIVE``).
        from_user: The user who sent the message.
        timestamp: The timestamp when the message was sent (in UTC).
        reply_to_message: The message to which this message is a reply to.
        body: The body of the message.
        token: The token of the flow. can be ``None`` in some cases :|
        response: The response from the flow.
        shared_data: Shared data between handlers.
    """


@dataclasses.dataclass(slots=True, kw_only=True)
class FlowDetails(_FlowDetails):
    """
    Represents the details of a flow.

    - Read more at `developers.facebook.com <https://developers.facebook.com/docs/whatsapp/flows/reference/flowsapi#details>`_.

    Attributes:
        id: The unique ID of the Flow.
        name: The user-defined name of the Flow which is not visible to users.
        status: The status of the flow.
        updated_at: The last time the flow was updated (name, categories, endpoint_uri, json, etc.).
        json_version: The version specified by the developer in the Flow JSON asset uploaded.
        data_api_version: The version of the Data API specified by the developer in the Flow JSON asset uploaded. Only for Flows with an Endpoint.
        categories: The categories of the flow.
        validation_errors: The validation errors of the flow (All errors must be fixed before the Flow can be published).
        endpoint_uri: The URL of the WA Flow Endpoint specified by the developer via API or in the Builder UI (Was ``data_channel_uri`` before v19.0).
        preview: The URL to the web preview page to visualize the flow and its expiry time.
        whatsapp_business_account: The WhatsApp Business Account which owns the Flow.
        application: The Facebook developer application used to create the Flow initially.
        health_status: A summary of the Flows health status.
    """

    _client: WhatsApp = dataclasses.field(repr=False, hash=False, compare=False)

    async def publish(self) -> SuccessResult:
        """
        Update the status of this flow to ``FlowStatus.PUBLISHED``.
            - A shortcut for :meth:`pywa.client.WhatsApp.publish_flow`.

        - This action is not reversible.
        - The Flow and its assets become immutable once published.
        - To update the Flow after that, you must create a new Flow. You specify the existing Flow ID as the clone_flow_id parameter while creating to copy the existing flow.

            You can publish your Flow once you have ensured that:

            - All validation errors and publishing checks have been resolved.
            - The Flow meets the design principles of WhatsApp Flows
            - The Flow complies with WhatsApp Terms of Service, the WhatsApp Business Messaging Policy and, if applicable, the WhatsApp Commerce Policy

        Returns:
            Whether the flow was published.

        Raises:
            FlowPublishingError: If this flow has validation errors or not all publishing checks have been resolved.
        """
        if res := await self._client.publish_flow(self.id):
            self.status = FlowStatus.PUBLISHED
        return res

    async def delete(self) -> SuccessResult:
        """
        When the flow is in ``FlowStatus.DRAFT`` status, you can delete it.
            - A shortcut for :meth:`pywa.client.WhatsApp.delete_flow`.

        Returns:
            Whether the flow was deleted.

        Raises:
            FlowDeletingError: If this flow is already published.
        """
        if res := await self._client.delete_flow(self.id):
            self.status = FlowStatus.DEPRECATED  # there is no `DELETED` status
        return res

    async def deprecate(self) -> SuccessResult:
        """
        When the flow is in ``FlowStatus.PUBLISHED`` status, you can only deprecate it.
            - A shortcut for :meth:`pywa.client.WhatsApp.deprecate_flow`.

        Returns:
            Whether the flow was deprecated.

        Raises:
            FlowDeprecatingError: If this flow is not published or already deprecated.
        """
        if res := await self._client.deprecate_flow(self.id):
            self.status = FlowStatus.DEPRECATED
        return res

    async def get_assets(self) -> Result[FlowAsset]:
        """
        Get all assets attached to this flow.
            - A shortcut for :meth:`pywa.client.WhatsApp.get_flow_assets`.

        Returns:
            The assets of the flow.
        """
        return await self._client.get_flow_assets(self.id)

    async def update_metadata(
        self,
        name: str | None = None,
        categories: Iterable[FlowCategory | str] | None = None,
        endpoint_uri: str | None = None,
        application_id: int | None = None,
    ) -> SuccessResult:
        """
        Update the metadata of this flow.
            - A shortcut for :meth:`pywa.client.WhatsApp.update_flow_metadata`.

        Args:
            name: The name of the flow (optional).
            categories: The new categories of the flow (optional).
            endpoint_uri: The URL of the FlowJSON Endpoint. Starting from FlowJSON 3.0 this property should be
             specified only gere. Do not provide this field if you are cloning a FlowJSON with version below 3.0.
            application_id: The ID of the Meta application which will be connected to the Flow. All the flows with endpoints need to have an Application connected to them.

        Example:

            >>> from pywa.types.flows import FlowCategory
            >>> wa = WhatsApp(business_account_id='1234567890', ...)
            >>> my_flows = wa.get_flows()
            >>> my_flows[0].update_metadata(
            ...     name='Feedback',
            ...     categories=[FlowCategory.SURVEY, FlowCategory.OTHER],
            ...     endpoint_uri='https://my-api-server/feedback_flow'
            ... )

        Returns:
            Whether the flow was updated.

        Raises:
            ValueError: If neither ``name``, ``categories`` or ``endpoint_uri`` is provided.
        """
        success = await self._client.update_flow_metadata(
            flow_id=self.id,
            name=name,
            categories=categories,
            endpoint_uri=endpoint_uri,
            application_id=application_id,
        )
        if success:
            if name:
                self.name = name
            if categories:
                self.categories = tuple(FlowCategory(c) for c in categories)
            if endpoint_uri:
                self.endpoint_uri = endpoint_uri
        return success

    async def update_json(
        self, flow_json: FlowJSON | dict | str | pathlib.Path | bytes | BinaryIO
    ) -> FlowJSONUpdateResult:
        """
        Update the json of this flow.
            - A shortcut for :meth:`pywa.client.WhatsApp.update_flow_json`.

        Args:
            flow_json: The new json of the flow. Can be a :class:`FlowJSON` object, :class:`dict`, json :class:`str`,
             json file path or json bytes.

        Returns:
            Whether the flow was updated.

        Raises:
            FlowUpdatingError: If the flow json is invalid or this flow is already published.
        """
        res = await self._client.update_flow_json(
            flow_id=self.id,
            flow_json=flow_json,
        )
        self.validation_errors = res.validation_errors or None
        return res



"""This module contains the types related to media."""

from __future__ import annotations


__all__ = [
    "Media",
    "Image",
    "Video",
    "Sticker",
    "Document",
    "Audio",
    "MediaUrlResponse",
]

import dataclasses
import mimetypes
from typing import TYPE_CHECKING

from .others import SuccessResult
from .. import utils

if TYPE_CHECKING:
    from ..client import WhatsApp


@dataclasses.dataclass(frozen=True, slots=True)
class Media:
    """Base class for all media types."""

    _client: WhatsApp = dataclasses.field(repr=False, hash=False, compare=False)
    id: str

    def get_media_url(self) -> str:
        """Gets the URL of the media. (expires after 5 minutes)"""
        return self._client.get_media_url(media_id=self.id).url

    def download(
        self,
        path: str | None = None,
        filename: str | None = None,
        in_memory: bool = False,
        **kwargs,
    ) -> bytes | str:
        """
        Download a media file from WhatsApp servers.
            - Same as :func:`~pywa.client.WhatsApp.download_media` with ``media_url=media.get_media_url()``

        >>> message.image.download()

        Args:
            path: The path where to save the file (if not provided, the current working directory will be used).
            filename: The name of the file (if not provided, it will be guessed from the URL + extension).
            in_memory: Whether to return the file as bytes instead of saving it to disk (default: False).
            **kwargs: Additional arguments to pass to ``httpx.get(...)``.

        Returns:
            The path of the saved file if ``in_memory`` is False, the file as bytes otherwise.
        """
        return self._client.download_media(
            url=self.get_media_url()
            if not hasattr(self, "url")
            else self.url,  # MediaUrlResponse
            path=path,
            filename=filename,
            in_memory=in_memory,
            **kwargs,
        )

    def delete(self, *, phone_id: str | int | None = utils.MISSING) -> SuccessResult:
        """
        Deletes the media from WhatsApp servers.

        Args:
            phone_id: The phone ID to delete the media from (optional, If included, the operation will only be processed if the ID matches the ID of the business phone number that the media was uploaded on. pass None to use the client's phone ID).
        """
        return self._client.delete_media(media_id=self.id, phone_id=phone_id)


@dataclasses.dataclass(frozen=True, slots=True, kw_only=True)
class BaseUserMedia(Media, utils.FromDict):
    """Base class for all media types."""

    sha256: str
    mime_type: str

    @property
    def extension(self) -> str | None:
        """Gets the extension of the media (with dot.)"""
        clean_mimetype = self.mime_type.split(";")[0].strip()
        return mimetypes.guess_extension(clean_mimetype)

    @classmethod
    def from_flow_completion(
        cls, client: WhatsApp, media: dict[str, str]
    ) -> BaseUserMedia:
        """
        Create a media object from the media dict returned by the flow completion.

        - You can use the shortcut :meth:`~pywa.types.FlowCompletion.get_media`

        Example:
            >>> from pywa import WhatsApp, types
            >>> wa = WhatsApp(...)
            >>> @wa.on_flow_completion
            ... def on_flow_completion(_: WhatsApp, flow: types.FlowCompletion):
            ...     img = types.Image.from_flow_completion(client=wa, media=flow.response['media'])
            ...     img.download()

        Args:
            client: The WhatsApp client.
            media: The media dict returned by the flow completion.

        Returns:
            The media object (Image, Video, Sticker, Document, Audio).
        """
        return cls.from_dict(media, _client=client)


@dataclasses.dataclass(frozen=True, slots=True, kw_only=True)
class Image(BaseUserMedia):
    """
    Represents an received image.

    Attributes:
        id: The ID of the file (can be used to download or re-send the image).
        sha256: The SHA256 hash of the image.
        mime_type: The MIME type of the image.
    """


@dataclasses.dataclass(frozen=True, slots=True, kw_only=True)
class Video(BaseUserMedia):
    """
    Represents a video.

    Attributes:
        id: The ID of the file (can be used to download or re-send the video).
        sha256: The SHA256 hash of the video.
        mime_type: The MIME type of the video.
    """


@dataclasses.dataclass(frozen=True, slots=True, kw_only=True)
class Sticker(BaseUserMedia):
    """
    Represents a sticker.

    Attributes:
        id: The ID of the file (can be used to download or re-send the sticker).
        sha256: The SHA256 hash of the sticker.
        mime_type: The MIME type of the sticker.
        animated: Whether the sticker is animated.
    """

    animated: bool


@dataclasses.dataclass(frozen=True, slots=True, kw_only=True)
class Document(BaseUserMedia):
    """
    Represents a document.

    Attributes:
        id: The ID of the file (can be used to download or re-send the document).
        sha256: The SHA256 hash of the document.
        mime_type: The MIME type of the document.
        filename: The filename of the document (optional).
    """

    filename: str | None = None


@dataclasses.dataclass(frozen=True, slots=True, kw_only=True)
class Audio(BaseUserMedia):
    """
    Represents an audio.

    Attributes:
        id: The ID of the file (can be used to download or re-send the audio).
        sha256: The SHA256 hash of the audio.
        mime_type: The MIME type of the audio.
        voice: Whether the audio is a voice message or just an audio file.
    """

    voice: bool


@dataclasses.dataclass(frozen=True, slots=True)
class MediaUrlResponse(Media, utils.FromDict):
    """
    Represents a media response.

    Attributes:
        id: The ID of the media.
        url: The URL of the media (valid for 5 minutes).
        mime_type: The MIME type of the media.
        sha256: The SHA256 hash of the media.
        file_size: The size of the media in bytes.
    """

    url: str
    file_size: int
    mime_type: str
    sha256: str



from __future__ import annotations

from typing import TYPE_CHECKING

"""This module contains types related to WhatsApp calls, including call connection, termination, and status updates."""


from pywa.types.calls import *  # noqa MUST BE IMPORTED FIRST
from pywa.types.calls import (
    CallConnect as _CallConnect,
    CallTerminate as _CallTerminate,
    CallStatus as _CallStatus,
    CallPermissionUpdate as _CallPermissionUpdate,
)

import dataclasses

from .base_update import BaseUserUpdateAsync  # noqa
from .callback import CallbackData
from .others import SuccessResult

if TYPE_CHECKING:
    from .sent_update import InitiatedCall
    from ..client import WhatsApp as WhatsAppAsync


class _CallShortcutsAsync:
    """Base class for async call actions."""

    id: str
    _client: WhatsAppAsync
    sender: str
    recipient: str

    async def pre_accept(self, *, sdp: SessionDescription) -> SuccessResult:
        """
        Pre-accept the call.

        - Read more at `developers.facebook.com <https://developers.facebook.com/docs/whatsapp/cloud-api/calling/user-initiated-calls#pre-accept-call>`_.

        In essence, when you pre-accept an inbound call, you are allowing the calling media connection to be established before attempting to send call media through the connection.

        When you then call the accept call endpoint, media begins flowing immediately since the connection has already been established

        Pre-accepting calls is recommended because it facilitates faster connection times and avoids `audio clipping issues <https://developers.facebook.com/docs/whatsapp/cloud-api/calling/troubleshooting#audio-clipping-issue-and-solution>`_.

        There is about 30 to 60 seconds after the Call Connect webhook is sent for the business to accept the phone call. If the business does not respond, the call is terminated on the WhatsApp user side with a "Not Answered" notification and a Terminate Webhook is delivered back to you.

        Args:
            sdp: Contains the session description protocol (SDP) type and description language.

        Returns:
            Whether the call was pre-accepted.
        """
        return await self._client.pre_accept_call(call_id=self.id, sdp=sdp)

    async def accept(
        self,
        sdp: SessionDescription,
        *,
        tracker: str | CallbackData | None = None,
    ) -> SuccessResult:
        """
        Connect to a call by providing a call agent's SDP.

        - Read more at `developers.facebook.com <https://developers.facebook.com/docs/whatsapp/cloud-api/calling/user-initiated-calls#accept-call>`_.

        You have about 30 to 60 seconds after the Call Connect Webhook is sent to accept the phone call. If your business does not respond, the call is terminated on the WhatsApp user side with a "Not Answered" notification and a Terminate Webhook is delivered back to you.

        Args:
            sdp: Contains the session description protocol (SDP) type and description language.
            tracker: The data to track the call with (optional, up to 512 characters, for complex data You can use :class:`CallbackData`).

        Returns:
            Whether the call was accepted.
        """
        return await self._client.accept_call(call_id=self.id, sdp=sdp, tracker=tracker)

    async def reject(self) -> SuccessResult:
        """
        Reject the call.

        - Read more at `developers.facebook.com <https://developers.facebook.com/docs/whatsapp/cloud-api/calling/user-initiated-calls#reject-call>`_.

        You have about 30 to 60 seconds after the Call Connect webhook is sent to accept the phone call. If the business does not respond the call is terminated on the WhatsApp user side with a "Not Answered" notification and a Terminate Webhook is delivered back to you.

        Returns:
            Whether the call was rejected.
        """
        return await self._client.reject_call(call_id=self.id)

    async def terminate(self) -> SuccessResult:
        """
        Terminate the active call.

        - Read more at `developers.facebook.com <https://developers.facebook.com/docs/whatsapp/cloud-api/calling/user-initiated-calls#terminate-call>`_.

        This must be done even if there is an RTCP BYE packet in the media path. Ending the call this way also ensures pricing is more accurate.
        When the WhatsApp user terminates the call, you do not have to call this endpoint. Once the call is successfully terminated, a Call Terminate Webhook will be sent to you.

        Returns:
            Whether the call was terminated.
        """
        return await self._client.terminate_call(call_id=self.id)


@dataclasses.dataclass(slots=True, kw_only=True, frozen=True)
class CallConnect(BaseUserUpdateAsync, _CallShortcutsAsync, _CallConnect):
    """
    Represents a call connection event.

    - This update arrives when a call is initiated by the business or the user.

    Attributes:
        id: The call ID.
        metadata: The metadata of the call (to which phone number this call was made or received).
        from_user: The user who participated in the call, either as caller or callee.
        timestamp: The timestamp when this call was made (in UTC).
        event: The calling event (always ``CONNECT``).
        direction: Whether the call was initiated by the business or the user.
        session: The session information, including SDP type and SDP info.
        shared_data: Shared data between handlers.
    """


@dataclasses.dataclass(slots=True, kw_only=True, frozen=True)
class CallTerminate(BaseUserUpdateAsync, _CallShortcutsAsync, _CallTerminate):
    """
    Represents a call termination event.

    - This update arrives when a call is terminated, either by the business or the user.

    Attributes:
        id: The call ID.
        metadata: The metadata of the call (to which phone number this call was made or received).
        from_user: The user who participated in the call, either as caller or callee.
        timestamp: The timestamp when this update is sent (in UTC).
        event: The calling event (always ``TERMINATE``).
        direction: Whether the call was initiated by the business or the user.
        status: The status of the call (either ``FAILED`` or ``COMPLETED``).
        start_time: The start time of the call in UTC (Only if the call was picked up).
        end_time: The end time of the call in UTC (Only if the call was picked up).
        duration: The duration of the call in seconds (Only if the call was picked up).
        tracker: The tracker that the call is initiated with.
        shared_data: Shared data between handlers.
    """

    async def recall(
        self, sdp: SessionDescription, *, tracker: str | CallbackData | None = None
    ) -> InitiatedCall:
        """
        Recall the call with the given SDP.

        - This is useful if you want to re-initiate a call after it has been terminated.

        Args:
            sdp: Contains the session description protocol (SDP) type and description language.
            tracker: The data to track the call with (optional, up to 512 characters, for complex data You can use :class:`CallbackData`).

        Returns:
            An InitiatedCall object containing the details of the initiated call.
        """
        return await self._client.initiate_call(
            to=self._internal_sender,
            sdp=sdp,
            tracker=tracker,
            phone_id=self._internal_recipient,
        )


@dataclasses.dataclass(slots=True, kw_only=True, frozen=True)
class CallStatus(BaseUserUpdateAsync, _CallShortcutsAsync, _CallStatus):
    """
    Represents a call status update.

    - This update arrives when during a business-initiated call, the user either accepts or rejects the call.

    Attributes:
        id: The call ID.
        metadata: The metadata of the call (to which phone number this call was made or received).
        timestamp: The timestamp when this status update is sent (in UTC).
        type: The type of the status update (always "call").
        status: The status of the call (either "RINGING", "ACCEPTED", or "REJECTED").
        tracker: The tracker that the call is initiated with.
        shared_data: Shared data between handlers.
    """


@dataclasses.dataclass(slots=True, kw_only=True, frozen=True)
class CallPermissionUpdate(BaseUserUpdateAsync, _CallPermissionUpdate):
    """
    Represents a call permission update.

    - This update arrives when a call permission request is sent to the user, and the user responds with an action.

    Attributes:
        id: The message ID.
        metadata: The metadata of the message (to which phone number this call permission request was sent).
        type: The type of the message (always ``INTERACTIVE``).
        reply_to_message: The message that this call permission request is replying to.
        from_user: The user who acted on the call permission request.
        timestamp: The timestamp when this update is sent (in UTC).
        response: The response to the call permission request (either ``ACCEPT`` or ``REJECT``).
        response_source: The source of the call permission response (either ``USER_ACTION`` or ``AUTOMATIC``).
        expiration_timestamp: The timestamp when the call permission request expires (if applicable).
    """



from __future__ import annotations

import functools
import time

from ..errors import WhatsAppError

"""Types for other objects."""

import dataclasses
import math
import logging
import datetime

from typing import (
    TYPE_CHECKING,
    Any,
    Iterable,
    TypeVar,
    Protocol,
    Generic,
    Iterator,
    ClassVar,
)

from .. import utils

if TYPE_CHECKING:
    from .message_status import MessageStatus
    from .chat_opened import ChatOpened
    from .media import Image, Video, Document, Audio, Sticker
    from .callback import CallbackButton, CallbackSelection
    from .calls import CallPermissions, CallingSettings
    from ..client import WhatsApp

_logger = logging.getLogger(__name__)


@dataclasses.dataclass(frozen=True, slots=True)
class User:
    """
    Represents a WhatsApp user.

    Attributes:
        wa_id: The WhatsApp ID of the user (The phone number with the country code).
        name: The name of the user (``None`` on :class:`MessageStatus` or when message type is :class:`MessageType.SYSTEM`).
        input: The input of the recipient is only available when sending a message.
    """

    _client: WhatsApp = dataclasses.field(repr=False, hash=False, compare=False)
    wa_id: str
    name: str | None
    input: str | None = dataclasses.field(
        default=None, repr=False, hash=False, compare=False
    )

    @classmethod
    def from_dict(cls, data: dict, client: WhatsApp) -> User:
        return cls(
            _client=client,
            wa_id=data["wa_id"],
            name=data.get("profile", {}).get("name"),
            input=data.get("input"),
        )

    def block(self) -> bool:
        """
        Block the user.

        - Shortcut for :meth:`~pywa.client.WhatsApp.block_users` with the user wa_id.

        Returns:
            bool: True if the user was blocked

        Raises:
            BlockUserError: If the user was not blocked
        """
        res = self._client.block_users((self.wa_id,))
        added = self.wa_id in {u.input for u in res.added_users}
        if not added:
            raise res.errors
        return added

    def unblock(self) -> bool:
        """
        Unblock the user.

        - Shortcut for :meth:`~pywa.client.WhatsApp.unblock_users` with the user wa_id.

        Returns:
            bool: True if the user was unblocked, False otherwise.
        """
        return self.wa_id in {
            u.input for u in self._client.unblock_users((self.wa_id,)).removed_users
        }

    def get_call_permissions(self) -> CallPermissions:
        """
        Get the call permissions of the user.

        - Shortcut for :meth:`~pywa.client.WhatsApp.get_call_permissions` with the user wa_id.

        Returns:
            CallPermissions: The call permissions of the user.
        """
        return self._client.get_call_permissions(wa_id=self.wa_id)

    def as_vcard(self) -> str:
        """Get the user as a vCard."""
        return "\n".join(
            (
                "BEGIN:VCARD",
                "VERSION:3.0",
                f"FN;CHARSET=UTF-8;ENCODING=QUOTED-PRINTABLE:{self.name}",
                f"TEL;type=CELL;type=VOICE:+{self.wa_id}",
                "END:VCARD",
            )
        )


class MessageType(utils.StrEnum):
    """
    Message types.

    Attributes:
        TEXT: Message.text -> :class:`str`.
        IMAGE: Message.image -> :class:`Image`.
        VIDEO: Message.video -> :class:`Video`.
        DOCUMENT: Message.document -> :class:`Document`.
        AUDIO: Message.audio -> :class:`Audio`.
        STICKER: Message.sticker -> :class:`Sticker`.
        REACTION: Message.reaction -> :class:`Reaction`.
        LOCATION: Message.location -> :class:`Location`.
        CONTACTS: Message.contacts -> tuple[:class:`Contact`].
        ORDER: Message.order -> :class:`Order`.
        UNKNOWN: An unknown message (Warning with the actual type will be logged).
        UNSUPPORTED: An unsupported message (message type not supported by WhatsApp Cloud API).
        INTERACTIVE: Only used in :class:`CallbackButton`, :class:`CallbackSelection` and :class:`CallPermissionUpdate`.
        BUTTON: Only used in :class:`CallbackButton`.
        REQUEST_WELCOME: Only used in :class:`ChatOpened`.
        SYSTEM: Only used in :class:`PhoneNumberChange` and :class:`IdentityChange`
    """

    _check_value = str.islower
    _modify_value = str.lower

    TEXT = "text"
    IMAGE = "image"
    VIDEO = "video"
    DOCUMENT = "document"
    AUDIO = "audio"
    STICKER = "sticker"
    REACTION = "reaction"
    LOCATION = "location"
    CONTACTS = "contacts"
    ORDER = "order"
    UNKNOWN = "unknown"
    UNSUPPORTED = "unsupported"

    INTERACTIVE = "interactive"
    BUTTON = "button"
    REQUEST_WELCOME = "request_welcome"
    SYSTEM = "system"


class InteractiveType(utils.StrEnum):
    """
    Interactive types.

    Attributes:

    """

    _check_value = str.islower
    _modify_value = str.lower

    BUTTON = "button"
    CTA_URL = "cta_url"
    CATALOG_MESSAGE = "catalog_message"
    LIST = "list"
    PRODUCT = "product"
    PRODUCT_LIST = "product_list"
    FLOW = "flow"
    LOCATION_REQUEST_MESSAGE = "location_request_message"
    VOICE_CALL = "voice_call"
    CALL_PERMISSION_REQUEST = "call_permission_request"

    UNKNOWN = "UNKNOWN"


@dataclasses.dataclass(frozen=True, slots=True)
class Reaction(utils.FromDict):
    """
    Represents a reaction to a message.

    Attributes:
        message_id: The ID of the message that was reacted to.
        emoji: The emoji that was used to react to the message (optional, ``None`` if removed).
    """

    message_id: str
    emoji: str | None = None

    @classmethod
    def from_dict(cls, data: dict, **kwargs) -> Reaction:
        return cls(
            message_id=data["message_id"], emoji=data.get("emoji") or None
        )  # sometimes it's empty string 🤦‍

    @property
    def is_removed(self) -> bool:
        """Check if the reaction is removed."""
        return self.emoji is None


@dataclasses.dataclass(frozen=True, slots=True)
class Location(utils.FromDict):
    """
    Represents a location.

    Attributes:
        latitude: The latitude of the location.
        longitude: The longitude of the location.
        name: The name of the location (optional).
        address: The address of the location (optional).
        url: The URL of the location (optional).
    """

    latitude: float
    longitude: float
    name: str | None = None
    address: str | None = None
    url: str | None = None

    @property
    def current_location(self) -> bool:
        """Check if the shared location is the current location or manually selected."""
        return not any((self.name, self.address, self.url))

    def in_radius(self, lat: float, lon: float, radius: float | int) -> bool:
        """
        Check if the location is in a radius of another location.

        Args:
            lat: The latitude of the other location.
            lon: The longitude of the other location.
            radius: The radius in kilometers.
        """
        lon1, lat1, lon2, lat2 = map(
            math.radians, [self.longitude, self.latitude, lon, lat]
        )
        return (
            (
                2
                * math.asin(
                    math.sqrt(
                        math.sin((lat2 - lat1) / 2) ** 2
                        + math.cos(lat1)
                        * math.cos(lat2)
                        * math.sin((lon2 - lon1) / 2) ** 2
                    )
                )
            )
            * 6371
        ) <= radius


@dataclasses.dataclass(slots=True)
class Contact:
    """
    Represents a contact.

    Attributes:
        name: The name of the contact.
        birthday: The birthday of the contact (in ``YYYY-MM-DD`` format, optional).
        phones: The phone numbers of the contact.
        emails: The email addresses of the contact.
        urls: The URLs of the contact.
        addresses: The addresses of the contact.
        org: The organization of the contact (optional).
    """

    name: Name
    birthday: str | None = None
    phones: Iterable[Phone] = dataclasses.field(default_factory=tuple)
    emails: Iterable[Email] = dataclasses.field(default_factory=tuple)
    urls: Iterable[Url] = dataclasses.field(default_factory=tuple)
    addresses: Iterable[Address] = dataclasses.field(default_factory=tuple)
    org: Org | None = None

    @classmethod
    def from_dict(cls, data: dict):
        return cls(
            name=cls.Name.from_dict(data["name"]),
            birthday=data.get("birthday"),
            phones=tuple(
                cls.Phone.from_dict(phone) for phone in data.get("phones", ())
            ),
            emails=tuple(
                cls.Email.from_dict(email) for email in data.get("emails", ())
            ),
            urls=tuple(cls.Url.from_dict(url) for url in data.get("urls", ())),
            addresses=tuple(
                cls.Address.from_dict(address) for address in data.get("addresses", ())
            ),
            org=cls.Org.from_dict(data["org"]) if "org" in data else None,
        )

    def to_dict(self) -> dict[str, Any]:
        """Get the contact as a dict."""
        return {
            "name": dataclasses.asdict(self.name),
            "birthday": self.birthday,
            "phones": tuple(dataclasses.asdict(phone) for phone in self.phones),
            "emails": tuple(dataclasses.asdict(email) for email in self.emails),
            "urls": tuple(dataclasses.asdict(url) for url in self.urls),
            "addresses": tuple(
                dataclasses.asdict(address) for address in self.addresses
            ),
            "org": dataclasses.asdict(self.org) if self.org else None,
        }

    def as_vcard(self) -> str:
        """Get the contact as a vCard."""
        return "\n".join(
            s
            for s in (
                "BEGIN:VCARD",
                "VERSION:3.0",
                f"FN;CHARSET=UTF-8;ENCODING=QUOTED-PRINTABLE:{self.name.formatted_name}",
                f"BDAY:{self.birthday}" if self.birthday else None,
                "\n".join(
                    f"TEL;type={phone.type}:{phone.phone}" for phone in self.phones
                )
                if self.phones
                else None,
                "\n".join(
                    f"EMAIL;type={email.type}:{email.email}" for email in self.emails
                )
                if self.emails
                else None,
                "\n".join(f"URL;type={url.type}:{url.url}" for url in self.urls)
                if self.urls
                else None,
                "\n".join(
                    f"ADR;type={a.type}:;;{';'.join((getattr(a, f) or '') for f in ('street', 'city', 'state', 'zip', 'country'))}"
                    for a in self.addresses
                )
                if self.addresses
                else None,
                "END:VCARD",
            )
            if s is not None
        )

    @dataclasses.dataclass(frozen=True, slots=True)
    class Name(utils.FromDict):
        """
        Represents a contact's name.

        - At least one of the optional parameters needs to be included along with the formatted_name parameter.

        Attributes:
            formatted_name: The formatted name of the contact.
            first_name: The first name of the contact (optional).
            last_name: The last name of the contact (optional).
            middle_name: The middle name of the contact (optional).
            suffix: The suffix of the contact (optional).
            prefix: The prefix of the contact (optional).
        """

        formatted_name: str
        first_name: str | None = None
        last_name: str | None = None
        middle_name: str | None = None
        suffix: str | None = None
        prefix: str | None = None

    @dataclasses.dataclass(frozen=True, slots=True)
    class Phone(utils.FromDict):
        """
        Represents a contact's phone number.

        Attributes:
            phone: The phone number (If ``wa_id`` is provided, No need for the ``phone``).
            type: The type of the phone number (Standard Values are CELL, MAIN, IPHONE, HOME, and WORK. optional).
            wa_id: The WhatsApp ID of the contact (optional).
        """

        phone: str | None = None
        type: str | None = None
        wa_id: str | None = None

    @dataclasses.dataclass(frozen=True, slots=True)
    class Email(utils.FromDict):
        """
        Represents a contact's email address.

        Attributes:
            email: The email address.
            type: The type of the email address (Standard Values are WORK and HOME. optional).
        """

        email: str | None = None
        type: str | None = None

    @dataclasses.dataclass(frozen=True, slots=True)
    class Url(utils.FromDict):
        """
        Represents a contact's URL.

        Attributes:
            url: The URL.
            type: The type of the URL (Standard Values are WORK and HOME. optional).
        """

        url: str | None = None
        type: str | None = None

    @dataclasses.dataclass(frozen=True, slots=True)
    class Org(utils.FromDict):
        """
        Represents a contact's organization.

        Attributes:
            company: The company of the contact (optional).
            department: The department of the contact (optional).
            title: The title of the business contact (optional).
        """

        company: str | None = None
        department: str | None = None
        title: str | None = None

    @dataclasses.dataclass(frozen=True, slots=True)
    class Address(utils.FromDict):
        """
        Represents a contact's address.

        Attributes:
            street: The street number and name of the address (optional).
            city: The city name of the address (optional).
            state: State abbreviation.
            zip: Zip code of the address (optional).
            country: Full country name.
            country_code: Two-letter country abbreviation (e.g. US, GB, IN. optional).
            type: The type of the address (Standard Values are WORK and HOME. optional).
        """

        street: str | None = None
        city: str | None = None
        state: str | None = None
        zip: str | None = None
        country: str | None = None
        country_code: str | None = None
        type: str | None = None


@dataclasses.dataclass(frozen=True, slots=True)
class ReferredProduct:
    """
    Represents a product this message is referring to.

    Attributes:
        catalog_id:
        sku: Unique identifier of the product in a catalog (also referred to as ``Content ID`` or ``Retailer ID``).

    """

    catalog_id: str
    sku: str

    @classmethod
    def from_dict(cls, data: dict) -> ReferredProduct:
        return cls(
            catalog_id=data["catalog_id"],
            sku=data["product_retailer_id"],
        )


@dataclasses.dataclass(frozen=True, slots=True)
class ReplyToMessage:
    """
    Represents a message that was replied to.

    Attributes:
        message_id: The ID of the message that was replied to.
        from_user_id: The ID of the user who sent the message that was replied to.
        referred_product: Referred product describing the product the user is requesting information about.
    """

    message_id: str
    from_user_id: str
    referred_product: ReferredProduct | None

    @classmethod
    def from_dict(cls, data: dict) -> ReplyToMessage:
        return cls(
            message_id=data["id"],
            from_user_id=data["from"],
            referred_product=ReferredProduct.from_dict(data["referred_product"])
            if "referred_product" in data
            else None,
        )


@dataclasses.dataclass(frozen=True, slots=True)
class Metadata(utils.FromDict):
    """
    Represents the metadata of a message.

    Attributes:
        display_phone_number: The phone number to which the message was sent.
        phone_number_id: The ID of the phone number to which the message was sent.
    """

    display_phone_number: str
    phone_number_id: str


@dataclasses.dataclass(frozen=True, slots=True)
class Product:
    """
    Represents a product in an order.

    Attributes:
        sku: Unique identifier of the product in a catalog (also referred to as ``Content ID`` or ``Retailer ID``).
        quantity: Number of items ordered.
        price: Price of the item.
        currency: Currency of the price.
    """

    sku: str
    quantity: int
    price: float
    currency: str

    @classmethod
    def from_dict(cls, data: dict) -> Product:
        return cls(
            sku=data["product_retailer_id"],
            quantity=data["quantity"],
            price=data["item_price"],
            currency=data["currency"],
        )

    @property
    def total_price(self) -> float:
        """Total price of the product."""
        return self.quantity * self.price


@dataclasses.dataclass(frozen=True, slots=True)
class Order:
    """
    Represents an order.

    Attributes:
        catalog_id: The ID for the catalog the ordered item belongs to.
        products:The ordered products.
        text: Text message from the user sent along with the order (optional).

    Properties:
        total_price: Total price of the order.
    """

    catalog_id: str
    products: tuple[Product, ...]
    text: str | None

    @classmethod
    def from_dict(cls, data: dict, _client: WhatsApp) -> Order:
        return cls(
            catalog_id=data["catalog_id"],
            text=data.get("text"),
            products=tuple(Product.from_dict(p) for p in data["product_items"]),
        )

    @property
    def total_price(self) -> float:
        """Total price of the order."""
        return sum(p.total_price for p in self.products)


@dataclasses.dataclass(frozen=True, slots=True)
class Referral(utils.FromDict):
    """
    Represents a referral object in a message.

    - This object is included in the messages object when a customer clicks an ad that redirects to WhatsApp.

    Attributes:
        source_url: The Meta URL that leads to the ad or post clicked by the customer.
        source_type: The type of the ad’s source; ``ad`` or ``post``.
        source_id: Meta ID for an ad or a post.
        headline: Headline used in the ad or post.
        body: Body for the ad or post.
        media_type: Media present in the ad or post; ``image`` or ``video``.
        image_url: URL of the image, when ``media_type`` is an ``image``.
        video_url: URL of the video, when ``media_type`` is a ``video``.
        thumbnail_url: URL for the thumbnail, when media_type is a video.
        ctwa_clid: Click ID generated by Meta for ads that click to WhatsApp.
    """

    source_url: str | None = None
    source_type: str | None = None
    source_id: str | None = None
    headline: str | None = None
    body: str | None = None
    media_type: str | None = None
    image_url: str | None = None
    video_url: str | None = None
    thumbnail_url: str | None = None
    ctwa_clid: str | None = None


@dataclasses.dataclass(slots=True)
class ProductsSection:
    """
    Represents a section in a section list.

    Attributes:
        title: The title of the products section (up to 24 characters).
        skus: The SKUs of the products in the section (at least 1, no more than 30).
    """

    title: str
    skus: Iterable[str]

    def to_dict(self) -> dict:
        return {
            "title": self.title,
            "product_items": tuple({"product_retailer_id": sku} for sku in self.skus),
        }


class Industry(utils.StrEnum):
    """
    Represents the industry of a business.

    Attributes:
        UNDEFINED: Undefined.
        OTHER: Other.
        AUTO: Automotive.
        BEAUTY: Beauty.
        APPAREL: Apparel.
        EDU: Education.
        ENTERTAIN: Entertainment.
        EVENT_PLAN: Event planning.
        FINANCE: Finance.
        GROCERY: Grocery store.
        GOVT: Government.
        HOTEL: Hotel.
        HEALTH: Health.
        NONPROFIT: Nonprofit.
        PROF_SERVICES: Professional services.
        RETAIL: Retail.
        TRAVEL: Travel.
        RESTAURANT: Restaurant.
        NOT_A_BIZ: Not a business.
    """

    UNDEFINED = "UNDEFINED"
    OTHER = "OTHER"
    AUTO = "AUTO"
    BEAUTY = "BEAUTY"
    APPAREL = "APPAREL"
    EDU = "EDU"
    ENTERTAIN = "ENTERTAIN"
    EVENT_PLAN = "EVENT_PLAN"
    FINANCE = "FINANCE"
    GROCERY = "GROCERY"
    GOVT = "GOVT"
    HOTEL = "HOTEL"
    HEALTH = "HEALTH"
    NONPROFIT = "NONPROFIT"
    PROF_SERVICES = "PROF_SERVICES"
    RETAIL = "RETAIL"
    TRAVEL = "TRAVEL"
    RESTAURANT = "RESTAURANT"
    NOT_A_BIZ = "NOT_A_BIZ"

    UNKNOWN = "UNKNOWN"


@dataclasses.dataclass(frozen=True, slots=True)
class BusinessProfile(utils.APIObject):
    """
    Represents a business profile.

    Attributes:
        about: This text appears in the business's profile, beneath its profile image, phone number, and contact buttons.
        address: Address of the business. Character limit 256.
        description: Description of the business. Character limit 512.
        email: The contact email address (in valid email format) of the business. Character limit 128.
        industry: The industry of the business.
        profile_picture_url: URL of the profile picture that was uploaded to Meta.
        websites: The URLs associated with the business. For instance, a website, Facebook Page, or Instagram.
         There is a maximum of 2 websites with a maximum of 256 characters each.
    """

    _override_api_fields: ClassVar = {"industry": "vertical"}

    about: str
    address: str | None
    industry: Industry
    description: str | None
    email: str | None
    profile_picture_url: str | None
    websites: tuple[str, ...] | None

    @classmethod
    def from_dict(cls, data: dict) -> BusinessProfile:
        return cls(
            about=data["about"],
            address=data.get("address"),
            industry=Industry(data["vertical"]),
            description=data.get("description"),
            email=data.get("email"),
            profile_picture_url=data.get("profile_picture_url"),
            websites=tuple(data.get("websites", ())) or None,
        )


@dataclasses.dataclass(frozen=True, slots=True)
class CommerceSettings(utils.APIObject):
    """
    Represents the WhatsApp commerce settings.

    Attributes:
        catalog_id: The ID of the catalog associated with the business.
        is_catalog_visible: Whether the catalog is visible to customers.
        is_cart_enabled: Whether the cart is enabled.
    """

    catalog_id: str
    is_catalog_visible: bool
    is_cart_enabled: bool

    @classmethod
    def from_dict(cls, data: dict) -> CommerceSettings:
        return cls(
            catalog_id=data["id"],
            is_catalog_visible=data["is_catalog_visible"],
            is_cart_enabled=data["is_cart_enabled"],
        )


class BusinessVerificationStatus(utils.StrEnum):
    """
    Represents the business verification status.

    Attributes:
        EXPIRED: The business verification has expired.
        FAILED: The business verification has failed.
        INELIGIBLE: The business is not eligible for verification.
        NOT_VERIFIED: The business is not verified.
        PENDING: The business verification is pending.
        PENDING_NEED_MORE_INFO: The business verification is pending and needs more information.
        PENDING_SUBMISSION: The business verification is pending submission.
        REJECTED: The business verification has been rejected.
        REVOKED: The business verification has been revoked.
        VERIFIED: The business is verified.
    """

    EXPIRED = "expired"
    FAILED = "failed"
    INELIGIBLE = "ineligible"
    NOT_VERIFIED = "not_verified"
    PENDING = "pending"
    PENDING_NEED_MORE_INFO = "pending_need_more_info"
    PENDING_SUBMISSION = "pending_submission"
    REJECTED = "rejected"
    REVOKED = "revoked"
    VERIFIED = "verified"

    UNKNOWN = "UNKNOWN"


class MarketingMessagesLiteAPIStatus(utils.StrEnum):
    """
    Represents the WhatsApp Business Account's status for onboarding onto Marketing Messages Lite.

    Attributes:
        INELIGIBLE: The WABA has not met all eligibility requirements.
        ELIGIBLE: The WABA has met all eligibility requirements.
        ONBOARDED: The WABA is eligible to use MM Lite API and has signed MM Lite Terms of Service, has a valid payment method attached, and is linked to an Ad account.
        UNKNOWN: An unknown status. Please contact support for assistance.
    """

    INELIGIBLE = "INELIGIBLE"
    ELIGIBLE = "ELIGIBLE"
    ONBOARDED = "ONBOARDED"
    UNKNOWN = "UNKNOWN"


class MarketingMessagesOnboardingStatus(utils.StrEnum):
    """
    Represents the WhatsApp Business Account's status for onboarding onto Marketing Messages Lite API.

    Attributes:
        INELIGIBLE_ON_BEHALF_OF_WABA: The WhatsApp Business Account (“WABA”) uses the “OBO” model which is not supported. You must either first transfer ownership of the WABA to the customer or onboard using the Intent API.
        INELIGIBLE_INACTIVE_OR_RESTRICTED: The WhatsApp Business Account is inactive or is restricted from messaging due to a policy enforcement issue.
        INELIGIBLE_COUNTRY_NOT_SUPPORTED: The WhatsApp Business Account is in a region that does not support the MM Lite API.
        INELIGIBLE_USING_WHATSAPP_BUSINESS_APP: The WhatsApp Business phone number is being used within the WhatsApp Business app.
        ELIGIBLE: WhatsApp Business Account is eligible to onboard and use the MM Lite API.
        PENDING_VALID_PAYMENT_METHOD: The WhatsApp Business Account requires setting up a valid payment method.
        PENDING_INTERNAL_SETUP: WhatsApp Business Account is in the configuration process of using the MM Lite API and requires no further action from the business customer or partner.
        ONBOARDED: The WhatsApp Business Account has successfully onboarded and is ready to use the MM Lite API.
    """

    INELIGIBLE_ON_BEHALF_OF_WABA = "INELIGIBLE_ON_BEHALF_OF_WABA"
    INELIGIBLE_INACTIVE_OR_RESTRICTED = "INELIGIBLE_INACTIVE_OR_RESTRICTED"
    INELIGIBLE_COUNTRY_NOT_SUPPORTED = "INELIGIBLE_COUNTRY_NOT_SUPPORTED"
    INELIGIBLE_USING_WHATSAPP_BUSINESS_APP = "INELIGIBLE_USING_WHATSAPP_BUSINESS_APP"
    ELIGIBLE = "ELIGIBLE"
    PENDING_VALID_PAYMENT_METHOD = "PENDING_VALID_PAYMENT_METHOD"
    PENDING_INTERNAL_SETUP = "PENDING_INTERNAL_SETUP"
    ONBOARDED = "ONBOARDED"

    UNKNOWN = "UNKNOWN"


@dataclasses.dataclass(frozen=True, slots=True)
class BusinessInfo(utils.APIObject, utils.FromDict):
    id: str
    name: str
    status: str
    type: str


@dataclasses.dataclass(frozen=True, slots=True)
class WhatsAppBusinessAccount(utils.APIObject):
    """
    Represents a WhatsApp Business Account.

    Attributes:
        id: The ID of the account.
        status: The status of the WhatsApp Business Account (e.g. ACTIVE).
        message_template_namespace: Namespace string for the message templates that belong to the WhatsApp Business Account
        name: User-friendly name to differentiate WhatsApp Business Accounts.
        timezone_id: The timezone of the WhatsApp Business Account (See `Timezone IDs <https://developers.facebook.com/docs/marketing-api/reference/ad-account/timezone-ids/>`_).
        business_verification_status: Current status of business verification of Meta Business Account which owns this WhatsApp Business Account
        is_enabled_for_insights: If true, indicates the WhatsApp Business Account enabled template analytics. See `Analytics <https://developers.facebook.com/docs/whatsapp/business-management-api/analytics>`_.
        marketing_messages_lite_api_status: WhatsApp Business Account's status for onboarding onto Marketing Messages Lite.
        marketing_messages_onboarding_status: Onboarding status of the WhatsApp Business account into Marketing Messages Lite API.
        ownership_type: Ownership type of the WhatsApp Business Account.
        currency: The currency in which the payment transactions for the WhatsApp Business Account will be processed
        country: country of the WhatsApp Business Account's owning Meta Business account

    """

    id: str
    name: str
    timezone_id: str
    message_template_namespace: str
    status: str | None
    business_verification_status: BusinessVerificationStatus | None
    is_enabled_for_insights: bool | None
    marketing_messages_lite_api_status: MarketingMessagesLiteAPIStatus | None
    marketing_messages_onboarding_status: MarketingMessagesOnboardingStatus | None
    on_behalf_of_business_info: BusinessInfo | None
    ownership_type: str | None
    health_status: dict | None
    currency: str | None
    country: str | None
    subscribed_apps: tuple[FacebookApplication, ...] | None

    @classmethod
    def from_dict(cls, data: dict) -> WhatsAppBusinessAccount:
        return cls(
            id=data["id"],
            status=data.get("status"),
            name=data["name"],
            timezone_id=data["timezone_id"],
            message_template_namespace=data.get("message_template_namespace"),
            business_verification_status=BusinessVerificationStatus(
                data["business_verification_status"]
            )
            if "business_verification_status" in data
            else None,
            is_enabled_for_insights=data.get("is_enabled_for_insights"),
            marketing_messages_lite_api_status=MarketingMessagesLiteAPIStatus(
                data["marketing_messages_lite_api_status"]
            )
            if "marketing_messages_lite_api_status" in data
            else None,
            marketing_messages_onboarding_status=MarketingMessagesOnboardingStatus(
                data["marketing_messages_onboarding_status"]
            )
            if "marketing_messages_onboarding_status" in data
            else None,
            on_behalf_of_business_info=BusinessInfo.from_dict(
                data["on_behalf_of_business_info"]
            )
            if "on_behalf_of_business_info" in data
            else None,
            ownership_type=data.get("ownership_type"),
            health_status=data.get("health_status"),
            currency=data.get("currency"),
            country=data.get("country"),
            subscribed_apps=tuple(
                FacebookApplication.from_dict(app["whatsapp_business_api_data"])
                for app in data["subscribed_apps"]["data"]
            )
            if "subscribed_apps" in data
            else None,
        )


@dataclasses.dataclass(frozen=True, slots=True)
class FacebookApplication(utils.FromDict, utils.APIObject):
    """
    Represents a Facebook Application.

    Attributes:
        id: The ID of the application.
        name: The name of the application.
        link: The link to the application.
    """

    id: str
    name: str
    link: str


class StorageStatus(utils.StrEnum):
    """
    Represents the storage status of a WhatsApp Business Phone Number.

    Attributes:
        DEFAULT: Default storage status.
        IN_COUNTRY_STORAGE_ENABLED: In-country storage is enabled.
    """

    DEFAULT = "DEFAULT"
    IN_COUNTRY_STORAGE_ENABLED = "IN_COUNTRY_STORAGE_ENABLED"

    UNKNOWN = "UNKNOWN"


@dataclasses.dataclass(slots=True, kw_only=True)
class StorageConfiguration:
    """
    Local storage offers an additional layer of data management control, by giving you the option to specify where your message data is stored at rest. If your company is in a regulated industry such as finance, government, or healthcare, you may prefer to have your message data stored in a specific country when at rest because of regulatory or company policies.

    Local storage is controlled by a setting enabled or disabled at a WhatsApp business phone number level. Both Cloud API and MM Lite API support local storage, and the setting will apply to any messages sent via either API if enabled.

    **How local storage works**

    When Local storage is enabled, the following constraints are applied to message content for a business phone number:

    - Data-in-use: When message content is sent or received by Cloud API or MM Lite API, message content may be stored on Meta data centers internationally while being processed.
    - Data-at-rest: After the data-in-use period, message content is deleted from Meta data centers outside of the specified local storage region, and persisted only in data centers within the local storage region selected. Note that the data-in-use period differs between Cloud API and MM Lite API as specified below:
        - When using local storage for Cloud API, the data-in-use period is up to 60 minutes.
        - When using local storage for MM Lite API, the data-in-use period is up to 90 minutes.
    The local storage feature supplements other WhatsApp Business Platform privacy and security controls, and allows customers to ensure a higher level of compliance with local data protection regulations.

    **Data in scope**

    Local storage applies to message content (text and media) sent and/or received via Cloud API and MM Lite API. The following message content are in scope of the local storage feature:

    - Text messages: text payload (message body)
    - Media messages: media payload (audio, document, image or video)
    - Template messages (static template + parameters passed at message send time): components with text / media payload
    In addition, a limited set of metadata attributes is included with the locally stored message content, in order to correctly associate the encrypted message payload with the originally processed message, and to audit the fact of localization. The stored metadata is protected with tokenization and encryption.

    **Available Regions**

    To see what regions are supported by local storage, see the ``data_localization_region`` parameter in the documentation on phone number registration.
    """

    status: str
    data_localization_region: str | None = None

    @classmethod
    def from_dict(cls, data: dict) -> StorageConfiguration:
        return cls(
            status=StorageStatus(data["status"]),
            data_localization_region=data.get("data_localization_region"),
        )


@dataclasses.dataclass(frozen=True, slots=True)
class Command:
    """
    Represents a command in a conversational automation.

    See `Conversational Automation <https://developers.facebook.com/docs/whatsapp/cloud-api/phone-numbers/conversational-components/#commands>`_.

    Attributes:
        name: The name of the command (without the slash).
        description: The description of the command.
    """

    name: str
    description: str

    @classmethod
    def from_dict(cls, data: dict):
        return cls(
            name=data.get("command_name"),
            description=data.get("command_description"),
        )

    def to_dict(self):
        return {
            "command_name": self.name,
            "command_description": self.description,
        }


@dataclasses.dataclass(frozen=True, slots=True)
class ConversationalAutomation:
    """
    Represents a conversational automation.

    See `Conversational Automation <https://developers.facebook.com/docs/whatsapp/cloud-api/phone-numbers/conversational-components>`_.

    Attributes:
        id: The ID of the WhatsApp Business Phone Number.
        chat_opened_enabled: Whether the welcome message is enabled (if so, you can listen to the :class:`ChatOpened` event).
        ice_breakers: See `Ice Breakers <https://developers.facebook.com/docs/whatsapp/cloud-api/phone-numbers/conversational-components/#ice-breakers>`_.
        commands: The `commands <https://developers.facebook.com/docs/whatsapp/cloud-api/phone-numbers/conversational-components/#commands>`_.
    """

    id: str
    chat_opened_enabled: bool
    ice_breakers: tuple[str] | None
    commands: tuple[Command, ...] | None

    @classmethod
    def from_dict(cls, data: dict):
        return cls(
            id=data.get("id"),
            chat_opened_enabled=data.get("enable_welcome_message", False),
            ice_breakers=tuple(data.get("prompts", ())) or None,
            commands=tuple(
                Command.from_dict(command) for command in data.get("commands", ())
            )
            or None,
        )


@dataclasses.dataclass(frozen=True, slots=True)
class BusinessPhoneNumber(utils.APIObject):
    """
    Represents a WhatsApp Business Phone Number.

    See `WhatsApp Business Phone Number <https://developers.facebook.com/docs/graph-api/reference/whats-app-business-account-to-number-current-status/>`_.

    Attributes:
        id: The ID of the phone number.
        verified_name: The name that appears in WhatsApp Manager and WhatsApp client chat thread headers,
         chat lists, and profile, if `display criteria <https://developers.facebook.com/docs/whatsapp/cloud-api/phone-numbers/#display-names>`_ is met.
        display_phone_number: International format representation of the phone number.
        conversational_automation: Conversational Automation feature config for this phone number.
        status: The operating status of the phone number (eg. connected, rate limited, warned).
        quality_rating: The quality rating of the phone number.
        quality_score: Quality score of the phone.
        webhook_configuration: The webhook configuration of the phone number.
        name_status: The status of the name review.
        new_name_status: The status of the review of the new name requested.
        code_verification_status: Indicates the phone number's one-time password (OTP) verification status. Values can be NOT_VERIFIED, VERIFIED, or EXPIRED.
         Only phone numbers with a VERIFIED status can be registered. See `Manage Phone Numbers and Certificates <https://developers.facebook.com/docs/whatsapp/embedded-signup/manage-accounts/phone-numbers#manage-phone-numbers-and-certificates>`_.
        account_mode: The account mode of the phone number. See `Filtering Phone Numbers <https://developers.facebook.com/docs/whatsapp/business-management-api/manage-phone-numbers#filter-phone-numbers>`_.
        is_on_biz_app: Indicates if the customer's business phone number is registered for WhatsApp Business app.
        is_official_business_account: Indicates if phone number is associated with an Official Business Account.
        is_pin_enabled: Returns True if a pin for two-step verification is enabled.
        is_preverified_number: Returns true if the phone number was pre-verified
        messaging_limit_tier: Current messaging limit tier.
        search_visibility: The availability of the phone_number in the WhatsApp Business search.
        platform_type: Platform the business phone number is registered with. Values can be CLOUD_API, ON_PREMISE, or NOT_APPLICABLE.
         If NOT_APPLICABLE, the number is not registered with Cloud API or On-Premises API.
        throughput: The business phone number's Cloud API throughput level. See `Phone Number Throughput <https://developers.facebook.com/docs/whatsapp/cloud-api/overview/#throughput>`_.
        eligibility_for_api_business_global_search: Status of eligibility in the API Business Global Search.
        health_status: health_status
        certificate: Certificate of the phone number
        new_certificate: Certificate of the new name that was requested
        last_onboarded_time: Indicates when the user added the business phone number to their WhatsApp Business Account
         (when the user completed the Embedded Signup flow).


    """

    id: str
    verified_name: str | None
    display_phone_number: str | None
    conversational_automation: ConversationalAutomation | None
    status: str | None
    quality_rating: str | None
    quality_score: dict[str, str] | None
    webhook_configuration: dict[str, str] | None
    name_status: str | None
    new_name_status: str | None
    code_verification_status: str | None
    account_mode: str | None
    is_on_biz_app: bool
    is_official_business_account: bool
    is_pin_enabled: bool
    is_preverified_number: bool
    messaging_limit_tier: str | None
    search_visibility: str | None
    platform_type: str | None
    throughput: dict[str, str] | None
    eligibility_for_api_business_global_search: str | None
    health_status: dict | None
    certificate: str | None
    new_certificate: str | None
    last_onboarded_time: str | None

    @classmethod
    def from_dict(cls, data: dict):
        return cls(
            id=data.get("id"),
            verified_name=data.get("verified_name"),
            display_phone_number=data.get("display_phone_number"),
            status=data.get("status"),
            conversational_automation=ConversationalAutomation.from_dict(
                data["conversational_automation"]
            )
            if data.get("conversational_automation")
            else None,
            quality_rating=data.get("quality_rating"),
            quality_score=data.get("quality_score"),
            webhook_configuration=data.get("webhook_configuration"),
            name_status=data.get("name_status"),
            new_name_status=data.get("new_name_status"),
            code_verification_status=data.get("code_verification_status"),
            account_mode=data.get("account_mode"),
            is_on_biz_app=data.get("is_on_biz_app"),
            is_official_business_account=data.get(
                "is_official_business_account", False
            ),
            is_pin_enabled=data.get("is_pin_enabled", False),
            is_preverified_number=data.get("is_preverified_number", False),
            messaging_limit_tier=data.get("messaging_limit_tier"),
            search_visibility=data.get("search_visibility"),
            platform_type=data.get("platform_type"),
            throughput=data.get("throughput"),
            eligibility_for_api_business_global_search=data.get(
                "eligibility_for_api_business_global_search"
            ),
            health_status=data.get("health_status"),
            certificate=data.get("certificate"),
            new_certificate=data.get("new_certificate"),
            last_onboarded_time=data.get("last_onboarded_time"),
        )


@dataclasses.dataclass(frozen=True, slots=True)
class QRCode(utils.APIObject):
    """
    Customers can scan a QR code from their phone to quickly begin a conversation with your business.
    The WhatsApp Business Management API allows you to create and access these QR codes and associated short links.

    Attributes:
        code: The code of the QR code.
        prefilled_message: The message that will be prefilled when the user starts a conversation with the business using the QR code.
        deep_link_url: The deep link URL of the QR code.
        qr_image_url: The URL of the QR code image (return only when creating a QR code).
    """

    code: str
    prefilled_message: str
    deep_link_url: str
    qr_image_url: str | None

    @classmethod
    def from_dict(cls, data: dict):
        return cls(
            code=data["code"],
            prefilled_message=data["prefilled_message"],
            deep_link_url=data["deep_link_url"],
            qr_image_url=data.get("qr_image_url"),
        )

    @classmethod
    @functools.cache
    def _api_fields(cls, image_type: str | None) -> tuple[str, ...]:
        fields = list(super(QRCode, cls)._api_fields())
        if image_type is not None:
            fields[fields.index("qr_image_url")] = f"qr_image_url.format({image_type})"
        else:
            fields.remove("qr_image_url")
        return tuple(fields)


@dataclasses.dataclass(slots=True, frozen=True)
class BlockUserFailure:
    """
    Represents a failure to block a user.

    Attributes:
        input: The phone number/wa_id input that failed to be blocked.
        errors: The errors that occurred during the operation.
    """

    input: str
    errors: tuple[WhatsAppError, ...]

    @classmethod
    def from_dict(cls, data: dict):
        return cls(
            input=data["input"],
            errors=tuple(WhatsAppError.from_dict(error) for error in data["errors"]),
        )


@dataclasses.dataclass(slots=True, frozen=True)
class UsersBlockedResult:
    """
    Represents the result of blocking users operation.

    Attributes:
        added_users: The users that were successfully blocked.
        failed_users: The users that failed to be blocked. You can access the .errors attribute in each failure to get the error details.
        errors: The errors that occurred during the operation (if any).
    """

    added_users: tuple[User, ...]
    failed_users: tuple[BlockUserFailure, ...]
    errors: WhatsAppError | None

    @classmethod
    def from_dict(cls, data: dict, client: WhatsApp):
        return cls(
            added_users=tuple(
                client._usr_cls.from_dict(user, client=client)
                for user in data.get("block_users", {}).get("added_users", [])
            ),
            failed_users=tuple(
                BlockUserFailure.from_dict(user)
                for user in data.get("block_users", {}).get("failed_users", [])
            ),
            errors=WhatsAppError.from_dict(data["errors"])
            if "errors" in data
            else None,
        )


@dataclasses.dataclass(frozen=True, slots=True)
class UsersUnblockedResult:
    """
    Represents the result of unblocking users operation.

    Attributes:
        removed_users: The users that were successfully unblocked.
    """

    removed_users: tuple[User, ...]

    @classmethod
    def from_dict(cls, data: dict, client: WhatsApp):
        return cls(
            removed_users=tuple(
                client._usr_cls.from_dict(user, client=client)
                for user in data.get("block_users", {}).get("removed_users", [])
            )
        )


@dataclasses.dataclass(slots=True, kw_only=True)
class Pagination:
    """
    Represents pagination parameters for fetching data.

    - See `Paginated Results <https://developers.facebook.com/docs/graph-api/results/>`_.

    **Cursor-based Pagination**

    Cursor-based pagination is the most efficient method of paging and should always be used when possible.
    A cursor refers to a random string of characters which marks a specific item in a list of data.
    The cursor will always point to the item, however it will be invalidated if the item is deleted or removed.
    Therefore, your app shouldn't store cursors or assume that they will be valid in the future.

    - Don't store cursors. Cursors can quickly become invalid if items are added or deleted.


    **Time-based Pagination**

    Time pagination is used to navigate through results data using Unix timestamps which point to specific times in a list of data.

    - For consistent results, specify both since and until parameters. Also, it is recommended that the time difference is a maximum of 6 months.

    **Offset-based Pagination**

    Offset pagination can be used when you do not care about chronology and just want a specific number of objects returned.
    Only use this if the edge does not support cursor or time-based pagination.
    Note that if new objects are added to the list of items being paged, the contents of each offset-based page will change.

    - Offset based pagination is not supported for all API calls. To get consistent results, we recommend you to paginate using the previous/next links we return in the response.


    Attributes:
        before: This is the cursor that points to the start of the page of data that has been returned.
        after: This is the cursor that points to the end of the page of data that has been returned.
        limit: This is the maximum number of objects that `may` be returned. A query may return fewer than the value of
         limit due to filtering. Do not depend on the number of results being fewer than the limit value to indicate
         that your query reached the end of the list of data, use the :attr:`~pywa.types.others.Result.has_next` instead as
         described below. For example, if you set limit to ``10`` and ``9`` results are returned, there may be more
         data available, but one item was removed due to privacy filtering. Some edges may also have a maximum on the
         limit value for performance reasons.
        offset: This offsets the start of each page by the number specified.
        until: A Unix timestamp or datetime obj value that points to the end of the range of time-based data.
        since: A Unix timestamp or datetime obj value that points to the start of the range of time-based data.
    """

    before: str | None = None
    after: str | None = None
    limit: int | None = None
    offset: int | None = None
    until: int | datetime.datetime | None = None
    since: int | datetime.datetime | None = None

    def to_dict(self) -> dict[str, Any]:
        """
        Convert the pagination parameters to a dictionary of request parameters.
        """
        params = {}
        if self.before:
            params["before"] = self.before
        if self.after:
            params["after"] = self.after
        if self.limit:
            params["limit"] = self.limit
        if self.offset:
            params["offset"] = self.offset
        if self.until:
            params["until"] = (
                int(self.until.timestamp())
                if isinstance(self.until, datetime.datetime)
                else self.until
            )
        if self.since:
            params["since"] = (
                int(self.since.timestamp())
                if isinstance(self.since, datetime.datetime)
                else self.since
            )
        return params


_T = TypeVar("_T")


class _ItemFactory(Protocol):
    def __call__(self, data: dict) -> _T: ...


class Result(Generic[_T]):
    """
    This class is used to handle paginated results from the WhatsApp API. You can iterate over the results, and also access the next and previous pages of results.

    - When using the ``next()`` or ``previous()`` methods, the results are returned as a new instance of the :class:`Result` class.
    - You can access the cursors using the ``before`` and ``after`` properties and use them later in the :class:`Pagination` object.

    Example:

        >>> from pywa import WhatsApp, types
        >>> wa = WhatsApp(...)
        >>> res = wa.get_blocked_users(pagination=types.Pagination(limit=100))
        >>> for user in res:
        ...     print(user.name, user.wa_id)
        ...
        >>> if res.has_next:
        ...     next_res = res.next()
        ...
        >>> print(res.all())

    Methods:
        next: Get the next page of results. if there is no next page, it returns empty Result.
        previous: Get the previous page of results. if there is no previous page, it returns empty Result.
        all: Get all results from the current page, previous pages, and next pages.
        empty: Returns an empty Result instance.

    Properties:
        has_next: Check if there is a next page of results.
        has_previous: Check if there is a previous page of results.
        before: Cursor that points to the start of the page of data that has been returned.
        after: Cursor that points to the end of the page of data that has been returned.
    """

    def __init__(
        self,
        wa: WhatsApp,
        response: dict,
        item_factory: _ItemFactory,
    ) -> None:
        self._wa = wa
        self._item_factory = item_factory
        self._data = [item_factory(item) for item in response.get("data", [])]
        self._next_url, self._previous_url = (
            response.get("paging", {}).get("next"),
            response.get("paging", {}).get("previous"),
        )
        self._cursors: dict = response.get("paging", {}).get("cursors", {})

    @property
    def has_next(self) -> bool:
        """Check if there is a next page of results."""
        return bool(self._next_url)

    @property
    def has_previous(self) -> bool:
        """Check if there is a previous page of results."""
        return bool(self._previous_url)

    @property
    def before(self) -> str | None:
        """Cursor that points to the start of the page of data that has been returned."""
        return self._cursors.get("before")

    @property
    def after(self) -> str | None:
        """Cursor that points to the end of the page of data that has been returned."""
        return self._cursors.get("after")

    @property
    def empty(self) -> Result[_T]:
        """Returns an empty Result instance."""
        return Result(
            wa=self._wa,
            response={
                "data": [],
                "paging": {"next": self._next_url, "cursors": self._cursors},
            },
            item_factory=self._item_factory,
        )

    def next(self) -> Result[_T]:
        """
        Get the next page of results. if there is no next page, it returns empty Result.

        - Check if there is a next page using the :attr:`~pywa.types.others.Result.has_next` property before calling this method.
        """
        if self.has_next:
            # noinspection PyProtectedMember
            response = self._wa.api._make_request(method="GET", endpoint=self._next_url)
            return self.__class__(
                wa=self._wa, response=response, item_factory=self._item_factory
            )
        return self.empty

    def previous(self) -> Result[_T]:
        """
        Get the previous page of results. if there is no previous page, it returns empty Result.

        - Check if there is a previous page using the :attr:`~pywa.types.others.Result.has_previous` property before calling this method.
        """
        if self.has_previous:
            # noinspection PyProtectedMember
            response = self._wa.api._make_request(
                method="GET", endpoint=self._previous_url
            )
            return self.__class__(
                wa=self._wa, response=response, item_factory=self._item_factory
            )
        return self.empty

    def all(
        self,
        *,
        sleep: float = 0.0,
    ) -> list[_T]:
        """
        Get all results from the current page, previous pages, and next pages.

        - Make sure to provide higher limit in the ``Pagination`` parameter to avoid hitting rate limits.
        - Also consider using the ``sleep`` parameter to avoid hitting rate limits.

        Args:
            sleep: The number of seconds to sleep between requests to avoid hitting rate limits. Default is 0.0 (no sleep).

        Returns:
            A list of all results from the current page, previous pages, and next pages.
        """
        before_data = []
        after_data = []

        prev = self
        while prev.has_previous:
            if sleep > 0:
                time.sleep(sleep)
            prev = prev.previous()
            before_data = prev._data + before_data

        next_page = self
        while next_page.has_next:
            if sleep > 0:
                time.sleep(sleep)
            next_page = next_page.next()
            after_data += next_page._data

        return before_data + self._data + after_data

    def __iter__(self) -> Iterator[_T]:
        return iter(self._data)

    def __len__(self) -> int:
        return len(self._data)

    def __getitem__(self, index: int) -> _T:
        return self._data[index]

    def __bool__(self) -> bool:
        return bool(self._data)

    def __repr__(self) -> str:
        return f"Result({self._data!r}, has_next={self.has_next!r}, has_previous={self.has_previous!r})"


@dataclasses.dataclass(frozen=True, slots=True)
class SuccessResult(utils.FromDict):
    """
    Represents a simple success result.

    - This is used for operations that do not return any data, but only indicate success or failure.

    You can use *this* class to check if an operation was successful or not::

        >>> wa = WhatsApp(...)
        >>> if wa.update_template(...): # update_template returns SuccessResult so we can check it directly
        ...     print("Template updated successfully")

    Attributes:
        success: Whether the operation was successful.
    """

    success: bool

    def __bool__(self) -> bool:
        """Returns True if the operation was successful."""
        return self.success



"""
Build WhatsApp Bots in Python • Fast, Effortless, Powerful 🚀

    - GitHub: https://github.com/david-lev/pywa
    - Docs: https://pywa.readthedocs.io/
    - PyPI: https://pypi.org/project/pywa/
"""

from pywa.client import WhatsApp
from pywa.utils import Version

__version__ = "3.0.0-rc.2"
__author__ = "David Lev"
__license__ = "MIT"



from typing import Any, Callable

from pywa.types import (
    MessageType,
    MessageStatusType,
    MarketingPreference,
)
from pywa.types.calls import (
    CallEvent,
    CallDirection,
    CallPermissionResponse,
    CallPermissionResponseSource,
)
from pywa.types.media import Image, Video, Document, Audio
from pywa.types.system import SystemType
from pywa.types.templates import TemplateStatus, TemplateCategory
from .common import CLIENTS

# {update_file: {update_name: [test_funcs]}}
TESTS: dict[str, dict[str, list[Callable[[Any], bool]]]] = {
    "message": {
        "text": [lambda m: m.text is not None],
        "image": [
            lambda m: m.image is not None,
            lambda m: m.has_media,
            lambda m: isinstance(m.media, Image),
        ],
        "video": [
            lambda m: m.video is not None,
            lambda m: m.has_media,
            lambda m: isinstance(m.media, Video),
        ],
        "document": [
            lambda m: m.document is not None,
            lambda m: m.has_media,
            lambda m: isinstance(m.media, Document),
        ],
        "audio": [
            lambda m: not m.audio.voice,
            lambda m: m.has_media,
            lambda m: isinstance(m.media, Audio),
        ],
        "voice": [lambda m: m.audio.voice],
        "static_sticker": [lambda m: not m.sticker.animated],
        "animated_sticker": [lambda m: m.sticker.animated],
        "reaction": [
            lambda m: m.reaction.emoji is not None,
            lambda m: m.id != m.message_id_to_reply,
        ],
        "unreaction_empty": [lambda m: m.reaction.emoji is None],
        "unreaction_no_emoji": [lambda m: m.reaction.emoji is None],
        "current_location": [lambda m: m.location.current_location],
        "chosen_location": [lambda m: not m.location.current_location],
        "contacts": [lambda m: m.contacts is not None],
        "order": [lambda m: m.order is not None],
        "unsupported": [lambda m: m.error is not None],
        "reply": [lambda m: m.is_reply],
        "forwarded": [lambda m: m.forwarded],
        "forwarded_many_times": [lambda m: m.forwarded_many_times],
        "interactive_message_with_err": [
            lambda m: m.type == MessageType.INTERACTIVE and m.error is not None
        ],
        "referral": [lambda m: m.referral is not None],
    },
    "callback_button": {
        "button": [lambda b: b.type == MessageType.INTERACTIVE],
        "quick_reply": [lambda b: b.type == MessageType.BUTTON],
    },
    "callback_selection": {
        "callback": [lambda s: s.data is not None],
        "description": [lambda s: s.description is not None],
    },
    "message_status": {
        "sent": [lambda s: s.status == MessageStatusType.SENT],
        "delivered": [lambda s: s.status == MessageStatusType.DELIVERED],
        "read": [lambda s: s.status == MessageStatusType.READ],
        "failed": [lambda s: s.error is not None],
        "with_tracker": [lambda s: s.tracker is not None],
    },
    "template_status_update": {
        "approved": [lambda s: s.new_status == TemplateStatus.APPROVED],
    },
    "template_category_update": {
        "marketing": [lambda c: c.new_category == TemplateCategory.MARKETING],
    },
    "template_components_update": {
        "element": [lambda c: c.template_element is not None],
    },
    "flow_completion": {
        "completion": [
            lambda f: f.token is not None,
            lambda f: "flow_token" not in f.response,
        ],
        "without_flow_token": [lambda f: f.token is None],
    },
    "chat_opened": {
        "chat_opened": [lambda c: c.type == MessageType.REQUEST_WELCOME],
    },
    "system": {
        "phone_number_change": [
            lambda s: s.sys_type == SystemType.USER_CHANGED_NUMBER,
            lambda s: s.sender == s.old_wa_id,
        ],
        "identity_change": [
            lambda s: s.sys_type == SystemType.CUSTOMER_IDENTITY_CHANGED,
        ],
    },
    "call_connect": {
        "business_initiated": [
            lambda c: c.direction == CallDirection.BUSINESS_INITIATED
        ],
        "user_initiated": [lambda c: c.direction == CallDirection.USER_INITIATED],
    },
    "call_terminate": {
        "call_terminate": [lambda c: c.event == CallEvent.TERMINATE],
    },
    "call_status": {
        "call_status": [lambda c: c.type is not None],
    },
    "call_permission_update": {
        "accept": [
            lambda c: c.response == CallPermissionResponse.ACCEPT
            and c.response_source == CallPermissionResponseSource.USER_ACTION
            and c.expiration_timestamp is not None,
        ],
        "reject": [
            lambda c: c.response == CallPermissionResponse.REJECT
            and c.response_source == CallPermissionResponseSource.USER_ACTION
            and c.expiration_timestamp is None,
        ],
    },
    "user_marketing_preferences": {
        "resume": [lambda u: u.value == MarketingPreference.RESUME],
    },
}


def test_types():
    for client, update_files in CLIENTS.items():
        for file, updates in update_files.items():
            for update_name, update in updates.items():
                for test_func in TESTS[file.stem][update_name]:
                    try:
                        assert test_func(update)
                    except AssertionError as e:
                        raise AssertionError(
                            f"Failed to assert update_name='{update_name}' error={e}"
                        )
